import os

import pytest
import yaml

from cnvrgv2.config import Config
from cnvrgv2.config import CONFIG_FOLDER_NAME
from cnvrgv2.data import FileDownloader
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.base.data_owner import DataOwner
from cnvrgv2.modules.project import Project
from cnvrgv2.proxy import Proxy


class MockResponse:
    def __init__(self, status, json):
        self.status_code = status
        self._json = json

    def json(self):
        return self._json


class MockFileHandler:
    def __init__(self):
        self.in_progress = False
        self.errors = []


class TestProject:
    def test_delete_project_with_slug(self, mocker, empty_context):
        slug = "test-slug"

        def mock_call_api_get(*args, **kwargs):
            return {
                "data": {
                    "id": "17",
                    "type": "project",
                    "attributes": {
                        "title": "test-title",
                        "slug": slug,
                        "git": False,
                        "start_commit": "b6692ea5df920cad691c20319a6fffd7a4a766b8",
                        "commit": "5b384ce32d8cdef02bc3a139d4cac0a22bb029e8"
                    }
                }
            }

        mocker.patch.object(Proxy, 'call_api', side_effect=mock_call_api_get)
        project = Project(context=empty_context, slug=slug)

        mocked_delete = mocker.patch.object(Proxy, 'call_api')
        project.delete()
        assert mocked_delete.call_count == 1

    def test_validate_config(self, mocker, empty_context):
        test_slug = "real_slug"

        def mock_local_config_file_fake_slug(*args, **kwargs):
            return {
                "commit_sha1": "test-commit",
                "git": True,
                "project_slug": "fake_slug",
            }

        def mock_local_config_file_real_slug(*args, **kwargs):
            return {
                "commit_sha1": "test-commit",
                "git": True,
                "project_slug": test_slug,
            }

        mocker.patch("builtins.open", return_value=True)
        mocker.patch.object(os.path, "exists", return_value=True)
        mocker.patch.object(Config, "_load_legacy_config", lambda x: True)

        mocker.patch.object(yaml, "safe_load", mock_local_config_file_fake_slug)
        project = Project(context=empty_context, slug=test_slug)
        assert not project._validate_config_ownership()

        mocker.patch.object(yaml, "safe_load", mock_local_config_file_real_slug)
        project = Project(context=empty_context, slug=test_slug)
        assert project._validate_config_ownership()

    def test_fail_put_files_with_illegal_directory_name(self, mocker, empty_context):
        test_slug = "test_slug"

        mocker.patch('os.path.isdir', return_value=True)

        project = Project(context=empty_context, slug=test_slug)
        with pytest.raises(CnvrgArgumentsError):
            project.put_files(['~/illegal-directory-name/file.txt'])

    def test_put_files_success(self, mocker, empty_context):
        test_slug = "test_slug"
        file_name = '~/directory/file.txt'
        expected_sha1 = 'sha1_mock'

        mocker.patch('os.path.isfile', return_value=True)
        mocker.patch('os.path.isdir', return_value=False)
        mocker.patch('os.path.getsize', return_value=1)

        project = Project(context=empty_context, slug=test_slug)

        start_commit_mock = mocker.patch.object(DataOwner, '_start_commit', return_value=expected_sha1)
        end_commit_mock = mocker.patch.object(DataOwner, '_end_commit', return_value=expected_sha1)
        file_uploader_mock = mocker.patch.object(DataOwner, '_get_file_uploader',
                                                 return_value=self._get_mock_file_handler())

        result = project.put_files(file_name)
        assert result == expected_sha1
        assert start_commit_mock.call_count == 1
        assert end_commit_mock.call_count == 1
        assert file_uploader_mock.call_count == 1

    def _get_mock_file_handler(self):
        return MockFileHandler()

    def _get_mock_file_handler_with_errors(self):
        mock = MockFileHandler()
        mock.errors = [Exception()]

    def test_put_git_files(self, mocker, empty_context):
        test_slug = "test_slug"
        file_name = '~/directory/file.txt'
        expected_sha1 = 'sha1_mock'

        mocker.patch('os.path.isfile', return_value=True)
        mocker.patch('os.path.isdir', return_value=False)
        mocker.patch('os.path.getsize', return_value=1)

        mocker.patch.object(DataOwner, '_handle_git_files', return_value=[file_name])

        start_commit_mock = mocker.patch.object(DataOwner, '_start_commit', return_value=expected_sha1)
        end_commit_mock = mocker.patch.object(DataOwner, '_end_commit', return_value=expected_sha1)
        file_uploader_mock = mocker.patch.object(DataOwner, '_get_file_uploader',
                                                 return_value=self._get_mock_file_handler())

        project = Project(slug=test_slug, context=empty_context)
        result = project.put_files([], git_diff=True)

        assert result == expected_sha1
        assert start_commit_mock.call_count == 1
        assert end_commit_mock.call_count == 1
        assert file_uploader_mock.call_count == 1

    def test_put_files_on_upload(self, mocker, empty_context):
        test_slug = "test_slug"
        file_name = '~/directory/file.txt'
        expected_sha1 = 'sha1_mock'

        mocker.patch('os.path.isfile', return_value=True)
        mocker.patch('os.path.isdir', return_value=False)
        mocker.patch('os.path.getsize', return_value=1)
        reload_mock = mocker.patch('cnvrgv2.modules.base.data_owner.DataOwner.reload')

        project = Project(context=empty_context, slug=test_slug)

        start_commit_mock = mocker.patch.object(DataOwner, '_start_commit', return_value=expected_sha1)
        end_commit_mock = mocker.patch.object(DataOwner, '_end_commit', return_value=expected_sha1)
        file_uploader_mock = mocker.patch.object(DataOwner, '_get_file_uploader',
                                                 return_value=self._get_mock_file_handler())
        file_deleter_mock = mocker.patch.object(DataOwner, '_get_remote_file_deleter',
                                                return_value=self._get_mock_file_handler())

        result = project.put_files(file_name, upload=True)
        assert result == expected_sha1
        assert start_commit_mock.call_count == 1
        assert end_commit_mock.call_count == 1
        assert file_uploader_mock.call_count == 1
        assert file_deleter_mock.call_count == 1
        assert reload_mock.call_count == 1

    def test_clone(self, mocker, empty_context, tmpdir):
        working_dir = "./"
        test_slug = "real_slug"
        reload_mock = mocker.patch('cnvrgv2.modules.base.data_owner.DataOwner.reload')

        def mock_file_downloader_init(obj, *args, **kwargs):
            # Set the credentials variables:
            obj.commit_sha1 = 'commit_sha1'
            obj.errors = []

        def mock_get_attr(attr):
            return 'mocked_{0}'.format(attr)

        # If project folder already exists, and override set to false, should not download any file
        # (assert there was no call to file downloader)
        with mocker.patch('os.path.exists', side_effect={working_dir + test_slug: True,
                                                         working_dir + test_slug + '/' + CONFIG_FOLDER_NAME: True}.get):
            mocked_file_downloader_init = mocker.patch.object(FileDownloader, "__init__",
                                                              side_effect=mock_file_downloader_init)
            project = Project(context=empty_context, slug=test_slug)
            mocker.patch('os.path.isfile', return_value=True)
            mocker.patch('os.path.isdir', return_value=False)
            assert project.local_commit is None
            project.clone(progress_bar_enabled=False, override=False, commit=None)
            assert reload_mock.call_count == 1
            assert mocked_file_downloader_init.call_count == 0
            assert project.local_commit is None

        # If project folder does not exist, download files
        with mocker.patch('os.path.exists', side_effect={working_dir + test_slug: True,
                                                         working_dir + test_slug + '/' + CONFIG_FOLDER_NAME: False}.get
                          ):
            mocker.patch.object(FileDownloader, "__init__", mock_file_downloader_init)
            mocker.patch.object(FileDownloader, "in_progress", new_callable=mocker.PropertyMock, return_value=False)
            save_config_mock = mocker.patch('cnvrgv2.modules.project.Project.save_config')
            create_cnvrg_ignore_mock = mocker.patch('cnvrgv2.modules.base.data_owner.create_cnvrgignore')
            mocker.patch.object(DataOwner, '__getattr__', side_effect=mock_get_attr)
            project = Project(context=empty_context, slug=test_slug)
            mocker.patch('os.path.isfile', return_value=True)
            mocker.patch('os.path.isdir', return_value=False)
            assert project.local_commit is None
            project.clone(progress_bar_enabled=False, override=False, commit=None)
            assert reload_mock.call_count == 3
            assert project.local_commit == 'mocked_last_commit'
            assert save_config_mock.call_count == 1
            assert create_cnvrg_ignore_mock.call_count == 1



from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.resources.templates.kube_template import KubeTemplate
from cnvrgv2.proxy import Proxy


class MockResponse:
    def __init__(self, status, json):
        self.status_code = status
        self._json = json

    def json(self):
        return self._json


class TestResource:
    def test_update_template_with_fake_attribute(self, mocker, empty_context, empty_cnvrg):

        empty_cnvrg._context = empty_context
        mocker.patch.object(Proxy, 'call_api', side_effect=MockResponse(status=200, json={}))
        template = KubeTemplate(context=empty_context, slug="slug", attributes={"is_private": False})
        try:
            template.update(fake_attribute=2.0)
        except CnvrgArgumentsError as err:
            assert "Bad arguments: Faulty key fake_attribute" in str(err)



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.labels.label import Label
from cnvrgv2.utils.json_api_format import JAF


class DataownerLabelsClient:
    def __init__(self, dataowner):
        self.dataowner = dataowner
        self._context = Context(context=dataowner._context)
        self._proxy = Proxy(context=self._context)

    def list(self, sort="-id"):
        """
        List all labels in a specific project/dataset
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields label objects
        """
        def label_builder(item):
            attributes = item.attributes
            attributes['id'] = item.id
            return Label(self._context, name=item.attributes['name'], kind=item.attributes['kind'], attributes=item.attributes)

        return api_list_generator(
            context=self._context,
            route=self._route,
            object_builder=label_builder,
            identifier='name',
            data={'kind': self.kind},
        )

    def remove(self):
        """
        Remove label from a specific project/dataset
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)

    def add(self, label):
        """
        Add label to a specific project/dataset
        """
        attributes = {
            "slug": self.dataowner.slug,
            "labels": [label.name],
        }

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="label", attributes=attributes)
        )



import os

import psutil
import pytest
import yaml

from cnvrgv2.config import Config, error_messages
from cnvrgv2.data import FileDownloader
from cnvrgv2.errors import CnvrgError, CnvrgNotEnoughSpaceError
from cnvrgv2.modules.commits.dataset_commit import DatasetCommit
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.proxy import Proxy
from cnvrgv2.utils.json_api_format import JAF


class TestDataset:
    def test_get_commit_valid_values(self, mocker, empty_context):
        dataset = Dataset(context=empty_context, slug="test-slug")

        commit = dataset.get_commit(sha_1="fake-sha1")
        assert type(commit) == DatasetCommit

    def test_get_commit_invalid_values(self, mocker, empty_context):
        dataset = Dataset(context=empty_context, slug="test-slug")

        with pytest.raises(CnvrgError):
            dataset.get_commit(sha_1=1)

    def test_dataset_as_request_params(self, mocker, empty_context):
        slug = "test-slug"

        mocker.patch.object(Proxy, 'call_api')

        dataset = Dataset(context=empty_context, slug=slug)
        as_params = dataset.as_request_params()
        assert as_params['slug']

    def test_delete_dataset(self, mocker, empty_context):
        slug = "test-slug"

        dataset = Dataset(context=empty_context, slug=slug)

        mocked_delete = mocker.patch.object(Proxy, 'call_api')
        dataset.delete()
        assert mocked_delete.call_count == 1

    def test_validate_config(self, mocker, empty_context):
        test_slug = "real_slug"

        def mock_local_config_file_fake_slug(*args, **kwargs):
            return {
                "commit_sha1": "test-commit",
                "git": True,
                "dataset_slug": "fake_slug",
            }

        def mock_local_config_file_real_slug(*args, **kwargs):
            return {
                "commit_sha1": "test-commit",
                "git": True,
                "dataset_slug": test_slug,
            }

        mocker.patch("builtins.open", return_value=True)
        mocker.patch.object(os.path, "exists", return_value=True)
        mocker.patch.object(Config, "_load_legacy_config", lambda x: True)

        mocker.patch.object(yaml, "safe_load", mock_local_config_file_fake_slug)
        dataset = Dataset(context=empty_context, slug=test_slug)
        assert not dataset._validate_config_ownership()

        mocker.patch.object(yaml, "safe_load", mock_local_config_file_real_slug)
        dataset = Dataset(context=empty_context, slug=test_slug)
        assert dataset._validate_config_ownership()

    def test_verify(self, mocker, empty_context):
        test_slug = "slug_1"
        dataset = Dataset(context=empty_context, slug=test_slug)

        def mock_config_init(obj):
            obj.local_config = {}

        # Mock after dataset creation, otherwise it will fail
        mocker.patch.object(os, "getcwd", return_value='test_dir')
        mocker.patch.object(os.path, "isdir", return_value=False)
        assert not dataset.verify()

        mocker.patch.object(os, "getcwd", return_value='test_dir')
        mocker.patch.object(os.path, "isdir", return_value=True)
        mocker.patch.object(os, "chdir", return_value=True)
        mocker.patch.object(Config, "__init__", mock_config_init)
        mocker.patch.object(Config, "local_config_exists", True)
        assert dataset.verify()

        mocker.patch.object(os, "getcwd", return_value='test_dir')
        mocker.patch.object(os.path, "isdir", return_value=True)
        mocker.patch.object(os, "chdir", return_value=True)
        mocker.patch.object(Config, "__init__", mock_config_init)
        mocker.patch.object(Config, "local_config_exists", False)
        assert not dataset.verify()

    def test_clone_dataset_disk_has_no_space_error(self, empty_context, mocker, tmpdir):
        dataset = Dataset(context=empty_context, slug="test-slug")
        mocker.patch('cnvrgv2.modules.base.data_owner.DataOwner.reload')

        def mock_file_downloader_init(obj, *args, **kwargs):
            obj.commit_sha1 = 'commit_sha1'
            obj.errors = []

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "dataset",
                    "attributes": {
                        "commit_size": 950,
                    }
                }
            }
            return JAF(response=response)

        mocker.patch.object(FileDownloader, "__init__", mock_file_downloader_init)
        mocker.patch.object(FileDownloader, "in_progress", new_callable=mocker.PropertyMock, return_value=False)
        mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        mocker.patch.object(psutil, 'disk_usage',
                            return_value=psutil._common.sdiskusage(free=650, used=0, total=0, percent=0))

        os.chdir(tmpdir.strpath)
        with pytest.raises(CnvrgNotEnoughSpaceError) as exception_info:
            dataset.clone(commit="test-commit")

        assert error_messages.NOT_ENOUGH_SPACE in str(exception_info.value)



from cnvrgv2.config import routes, error_messages
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.resources.templates.spark_templates_client import SparkTemplatesClient
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes


class SparkDriver(DynamicAttributes):
    available_attributes = {
        # TODO: Add relevant attributes
        "slug": str,
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current spark driver
        if slug:
            self._context.set_scope(SCOPE.RESOURCE, slug)

        scope = self._context.get_scope(SCOPE.RESOURCE)

        self._proxy = Proxy(context=self._context)
        self._route = routes.SPARK_DRIVER_BASE.format(scope["organization"], scope["resource"])
        self._attributes = attributes or {}
        self.slug = scope["resource"]
        self._init_clients()

    def save(self):
        pass

    def delete(self):
        """
        Deletes the current spark driver
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)

    def _init_clients(self):
        """
        Sets up the clients that are exposed to user user via machine
        @return: Void
        """
        try:
            self.templates = SparkTemplatesClient(context=self._context)
        except CnvrgArgumentsError:
            raise CnvrgArgumentsError(error_messages.FAILED_TO_INIT_CLIENT.format('templates'))



import re
import click
from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from functools import wraps


def shared_options(**kwargs):
    def inner_decorator(func):
        @click.option('-mdw', '--max-duration-workspaces', type=int, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_MAX_DURATION_WORKSPACES, **kwargs)
        @click.option('-mde', '--max-duration-experiments', type=int, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_MAX_DURATION_EXPERIMENTS, **kwargs)
        @click.option('-mde', '--max-duration-endpoints', type=int, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_MAX_DURATION_ENDPOINTS, **kwargs)
        @click.option('-mdw', '--max-duration-webapps', type=int, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_MAX_DURATION_WEBAPPS, **kwargs)
        @click.option('-accc', '--automatically-clear-cached-commits', type=int, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_AUTOMATICALLY_CLEAR_CACHED_COMMITS, **kwargs)
        @click.option('-swu', '--slack-webhook-url', type=str, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_SLACK_WEBHOOK_URL, **kwargs)
        @click.option('-qcwt', '--queued-compute-wait-time', type=int, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_QUEUED_COMPUTE_WAIT_TIME, **kwargs)
        @click.option('-cpu', '--custom-pypi-url', type=str, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_CUSTOM_PYPI_URL, **kwargs)
        @click.option('-dt', '--debug-time', type=int, default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_DEBUG_TIME, **kwargs)
        @click.option('-it', '--idle-time', type=int, default=None, help=messages.ORGANIZATION_SETTINGS_HELP_IDLE_TIME,
                      **kwargs)
        @click.option('-ie', '--idle-enabled', default=None, help=messages.ORGANIZATION_SETTINGS_HELP_IDLE_ENABLED,
                      is_flag=True)
        @click.option('-dte', '--debug-time-enabled', default=None, help=messages.ORGANIZATION_SETTINGS_HELP_DEBUG_TIME,
                      is_flag=True)
        @click.option('-eoe', '--email-on-error', default=None, help=messages.ORGANIZATION_SETTINGS_HELP_EMAIL_ON_ERROR,
                      is_flag=True)
        @click.option('-eos', '--email-on-success', default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_EMAIL_ON_SUCCESS, is_flag=True)
        @click.option('-cpe', '--custom-pypi-enabled', default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_CUSTOM_PYPI_ENABLED, is_flag=True)
        @click.option('-id', '--install-dependencies', default=None,
                      help=messages.ORGANIZATION_SETTINGS_HELP_INSTALL_DEPENDENCIES, is_flag=True)
        @wraps(func)
        def wrapper(*args, **_kwargs):
            return func(*args, **_kwargs)

        return wrapper

    return inner_decorator


@click.group(name="settings")
def organization_settings_group():
    pass


@organization_settings_group.command()
@prepare_command()
def show(cnvrg, logger):
    click.echo("Current settings for organization:")
    cnvrg.settings.reload()
    for prop, value in cnvrg.settings.values().items():
        click.echo(prop + ": " + str(value))


@organization_settings_group.command(name="set")
@shared_options()
@click.option('-dc', '--default-computes', type=str, default=None,
              help=messages.ORGANIZATION_SETTINGS_HELP_DEFAULT_COMPUTES)
@prepare_command()
def set_settings(cnvrg, logger, default_computes, **kwargs):
    if default_computes is not None:
        cnvrg.settings.update(default_computes=re.split(r'\s*,\s*', default_computes))
        return

    cnvrg.settings.update(**{k: v for k, v in kwargs.items() if v is not None})


@organization_settings_group.command(name="unset")
@shared_options(is_flag=True)
@prepare_command()
def unset_settings(
        cnvrg,
        logger,
        **kwargs
):
    cnvrg.settings.update(**{k: None for k, v in kwargs.items() if v is not None})



import os

import click

from cnvrgv2.cli.logger.logger import CnvrgLogger


@click.command()
def logs():
    """
    Print last 50 log files
    """
    log_file_path = os.path.join(CnvrgLogger.LOGS_DIR, CnvrgLogger.LOG_FILE_NAME)
    try:
        with open(log_file_path, 'r') as f:
            lines = f.readlines()
            if len(lines) > 50:
                lines = lines[-50:]
            for line in lines:
                click.echo(line)

        click.echo(click.style("Full logs can be seen at {}".format(CnvrgLogger.LOGS_DIR), fg="magenta", italic=True))

    except FileNotFoundError:
        click.echo("There are currently no logs available. Logs will be generated after using cli commands")



import numbers

from cnvrgv2.config.error_messages import BOTH_GROUP_AND_STEP, ONLY_HEATMAP_CAN_HAVE_GROUPS, SERIES_NOT_A_HEATMAP, \
    SERIES_NOT_A_LIST, SERIES_NOT_A_VALID_LIST
from cnvrgv2.errors import CnvrgArgumentsError


# Chart types ENUM
class CHART_TYPES:
    LINE = "none"
    SCATTER = "scatter"
    BAR = "bar"
    HEATMAP = "heatmap"


# Helpers


# To use numpy lists and such without errors
def sanitize_list(lst):
    if hasattr(lst, 'tolist'):
        return lst.tolist()
    return lst


def create_series_data(Ys: list):
    y_values = sanitize_list(Ys)

    # If no data was sent, do nothing.
    if not y_values:
        return None
    return y_values


def check_heatmap_series(series, matrix_size):
    # Verifying data format is heatmap - tuples of the same length made of real numbers
    # for ex. [(1,2), (1,2)], [(1,2,3), (1,2,3)], etc.

    # Skip verifications on empty series
    if len(series) == 0:
        return

    if not isinstance(series[0], tuple):
        raise CnvrgArgumentsError(SERIES_NOT_A_HEATMAP)

    value_length = matrix_size

    for value in series:
        if (
            not isinstance(value, tuple) or
            not len(value) == value_length or
            not all(isinstance(n, numbers.Real) for n in value)
        ):
            raise CnvrgArgumentsError(SERIES_NOT_A_HEATMAP)


# Charts
class Chart:
    def __init__(self, key, chart_type, **kwargs):
        """
        Represents a chart
        @param key: Identifier of the chart.
        @param chart_type: Type of the chart, from CHART_TYPES Enum.
        @param kwargs: additional arguments:
            title: chart title,
            group: chart group,
            step: data steps,
            x_axis: X axis custom title,
            y_axis: Y axis custom title,
            Ys / y_ticks: Y axis data, this is optional as user can add serieses later.
            Xs / x_ticks / bars: X axis custom ticks (X values that will correspond to Y values sent / bar names),
            series_name: Name of the initial series to be built (the series that's based on Ys & Xs)
            stops: list of color stops,
            max_val: max boundary for Y axis,
            min_val: min boundary for Y axis,
            colors: custom colors to use on each series
        """
        self.series_list = []
        self.chart_type = chart_type
        self.key = key
        self.kwargs = {
            **kwargs,
            'x_ticks': kwargs.get('Xs', None) or kwargs.get('bars', None) or kwargs.get('x_ticks', None),
            'min': kwargs.get('min_val', None),
            'max': kwargs.get('max_val', None)
        }

        # Make sure only heatmap can get group and step
        if chart_type != CHART_TYPES.HEATMAP and (self.kwargs.get('step', None) or self.kwargs.get('group', None)):
            raise CnvrgArgumentsError(ONLY_HEATMAP_CAN_HAVE_GROUPS)

        # Values for initial series or heatmap. This is optional. Users can add data manually later.
        init_matrix = self.kwargs.get('matrix', None)
        init_series_name = self.kwargs.get('series_name', None)
        if init_matrix:
            self.add_series(data=init_matrix, name=init_series_name)
        else:
            init_series_ys = self.kwargs.get('Ys', None)
            init_series_data = create_series_data(Ys=init_series_ys)
            self.add_series(data=init_series_data, name=init_series_name)

    def add_series(self, data, name=None):
        """
        Adds series to the chart
        @param data: list of data points.
        In case of regular charts, will be a list of numbers or Point objects.
        In case of bar charts, will only be a list of numbers.
        In case of Heatmap, list of tuples that form a matrix. e.g, 3x3 matrix: [(1,2,3),(4,5,6),(7,8,9)]
        @param name: name of the series
        @return: None
        """
        if not data:
            return

        formatted_data = self.validate_series(data)

        # If series already exists add new data to that series, else add the series
        existing_series = list(map(lambda series: series.get('name', None), self.series_list))
        if name in existing_series:
            index, current_series = next(
                (i, series) for (i, series) in enumerate(self.series_list) if series["name"] == name
            )

            current_series["data"].extend(formatted_data)
            self.series_list[index] = current_series
        else:
            self.series_list.append({
                "name": name,
                "data": formatted_data
            })

    def to_dict(self):
        """
        @return: The chart's attributes as dictionary
        """
        return {
            "chart_type": self.chart_type,
            "key": self.key,
            "series": self.series_list,
            **self.kwargs
        }

    def validate_series(self, series):
        """
        Validates series before adding to series_list.
        In base class, only verifies series is made of real numbers
        @param series: Series to convert
        @return: The series converted
        """
        if not isinstance(series, list):
            raise CnvrgArgumentsError(SERIES_NOT_A_LIST)

        if not all(isinstance(value, numbers.Real) for value in series):
            raise CnvrgArgumentsError(SERIES_NOT_A_VALID_LIST)
        return series


class Heatmap(Chart):
    def __init__(self, key, **kwargs):
        # Set Ys as y_ticks so user can use both
        y_ticks = kwargs.get('Ys', None) or kwargs.get('y_ticks', None)
        self.matrix_size = kwargs.get('matrix_size', None)
        self.matrix_last_pos = kwargs.get('matrix_last_pos', 0)
        formatted_kwargs = {**kwargs, 'y_ticks': y_ticks}

        # Validate group and step are always both available
        self.group = kwargs.get('group', None)
        self.step = kwargs.get('step', None)
        if (self.group and not self.step) or (self.step and not self.group):
            raise CnvrgArgumentsError(BOTH_GROUP_AND_STEP)

        super().__init__(key=key, chart_type=CHART_TYPES.HEATMAP, **formatted_kwargs)

    def _convert_series(self, series):
        matrix = []
        for i, y in enumerate(range(self.matrix_last_pos, self.matrix_last_pos + len(series))):
            for x in range(len(series[0])):
                matrix.append((x, y, series[i][x]))
        return matrix

    def validate_series(self, series):
        """
        Validates and converts series to heatmap format before adding to series_list.
        @param series: Series to convert
        @return: The series converted
        """

        if not isinstance(series, list):
            raise CnvrgArgumentsError(SERIES_NOT_A_LIST)

        # This means this a new matrix being created
        if not (self.matrix_size and self.matrix_size > 1):
            self.matrix_size = len(series[0])

        check_heatmap_series(series, self.matrix_size)
        return self._convert_series(series)


class BarChart(Chart):
    def __init__(self, key, **kwargs):
        super().__init__(key=key, chart_type=CHART_TYPES.BAR, **kwargs)


class ScatterPlot(Chart):
    def __init__(self, key, **kwargs):
        super().__init__(key=key, chart_type=CHART_TYPES.SCATTER, **kwargs)


class LineChart(Chart):
    def __init__(self, key, **kwargs):
        super().__init__(key=key, chart_type=CHART_TYPES.LINE, **kwargs)



import os

import yaml

from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgError
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.modules.resources.templates.kube_templates_client import KubeTemplatesClient
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.url_utils import urljoin


class Cluster(DynamicAttributes):
    available_attributes = {
        "slug": str,
        "title": str,
        "status": str,
        "is_default": bool,
        "cluster_type": str,
        "domain": str,
        "last_health_check": str,
        "scheduler": str,
        "https_scheme": bool,
        "namespace": str,
        "persistent_volumes": bool,
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current machine
        if slug:
            self._context.set_scope(SCOPE.RESOURCE, slug)

        scope = self._context.get_scope(SCOPE.RESOURCE)

        self._proxy = Proxy(context=self._context)
        self._route = routes.CLUSTER_BASE.format(scope["organization"], scope["resource"])
        self._attributes = attributes or {}
        self.slug = scope["resource"]
        self._init_clients()

    def save(self):
        pass

    def delete(self):
        """
        Deletes the current cluster
        @return: None
        """
        # TODO: change routes without resource_request
        scope = self._context.get_scope(SCOPE.RESOURCE)
        base_route = routes.CLUSTERS_BASE.format(scope["organization"])
        route = urljoin(base_route, routes.CLUSTER_RESOURCE_REQUEST, self.slug)
        self._proxy.call_api(route=route, http_method=HTTP.DELETE)

    def _init_clients(self):
        """
        Sets up the clients that are exposed to user user via machine
        @return: Void
        """
        try:
            self.templates = KubeTemplatesClient(context=self._context)
        except CnvrgError:
            # TODO: Surpress exceptions here
            pass

    def update(self,
               kube_config_yaml_path=None,
               resource_name=None,
               **kwargs
               ):
        """
        @param kube_config_yaml_path: path to the kube_config yaml file
        @param resource_name: the name of the resource
        @param kwargs: kwargs: Dictionary. attributes for updating
            for existing kubernetes_cluster:
                scheduler: cnvrg_scheduler
                namespace: cnvrg
                domain: domain url of existing kubernetes_cluster
                https_scheme: false
                persistent_volumes: false
                gaudi_enabled: false
        @return: None
        """
        if kube_config_yaml_path and not os.path.exists(kube_config_yaml_path):
            raise CnvrgArgumentsError(error_messages.NO_FILE)

        attributes = {**kwargs}
        if kube_config_yaml_path:
            kube_config_yaml_content = yaml.safe_load(open(kube_config_yaml_path, "r"))
            attributes["kube_config"] = str(kube_config_yaml_content)

        if resource_name:
            attributes["name"] = resource_name

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=attributes
        )



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.validators import attributes_validator, validate_gpu, validate_template_type
from cnvrgv2.modules.resources.templates.kube_template import KubeTemplate
from cnvrgv2.utils.casters import safe_cast

required_attributes = [
    'title',
    'parent_cpu',
    'parent_memory',
]


class KubeTemplatesClient:
    def __init__(self, context=None):
        self._context = Context(context=context)
        scope = self._context.get_scope(SCOPE.RESOURCE)
        self._proxy = Proxy(context=self._context)
        self._route = routes.TEMPLATES_BASE.format(scope["organization"], 'clusters', scope['resource'])

    def list(self, sort="-id"):
        """
        List all kube compute templates in a specific resource
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields kube compute templates objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=KubeTemplate,
            sort=sort,
        )

    def get(self, slug):
        """
        Retrieves a kube compute template by the given slug
        @param slug: The slug of the requested kube compute template
        @return: kube compute template object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.TEMPLATE_GET_FAULTY_SLUG)

        return KubeTemplate(context=self._context, slug=slug)

    def create(
            self,
            title,
            cpu,
            memory,
            **kwargs
    ):
        """

        @param title: Name of the new kube template
        @param cpu: CPU cores of the new kube template
        @param memory: RAM memory of the new kube template
        @param kwargs: A list of optional attributes for creation
            gpu: GPU cores of the new kube template
            gaudi: Gaudi accelerators num of the new kube template
            huge_pages: Gaudi huge pages count of the kube template
            is_private: is the new kube template private or public
            users: Users with access to the new kube template
            mig_device: mig device type for the new kube template
            shm: Shared memory (RAM) count of the kube template
            template_type: template type of the new kube template
            labels: labels of the new kube template
            taints: taints of the new kube template
            worker_num_executors: num of workers for the new kube template
            worker_cpu: worker CPU cores of the new kube template
            worker_memory: worker RAM memory of the new kube template
            worker_gpu: worker GPU cores of the new kube template
            worker_mig_device: worker mig device type
            worker_labels: worker labels
            worker_taints: worker taints

        @return: The newly created kube template
        """

        formatted_attributes = {
            "title": title,
            "type": kwargs.get("template_type") or 'regular',
            "parent_cpu": float(cpu),
            "parent_memory": float(memory),
            "parent_labels": kwargs.get("labels"),
            "parent_taints": kwargs.get("taints"),
            "parent_shm": safe_cast(kwargs.get("shm"), float),
            "parent_gaudi": safe_cast(kwargs.get("gaudi"), float),
            "parent_huge_pages": safe_cast(kwargs.get("huge_pages"), float),
            **kwargs,
            "parent_gpu": {
                "count": safe_cast(kwargs.get("gpu"), float),
                "mig_device": kwargs.get("mig_device")
            },
            "worker_gpu": {
                "count": safe_cast(kwargs.get("worker_gpu"), float),
                "mig_device": kwargs.get("worker_mig_device"),
            }
        }
        # The cnvrg api requires cpu and memory of the parent, For other templates - parent and/or worker.
        # for pytorch template - we need to define only worker, but only worker is not valid.
        # so we convert parent into worker and send it to the api.
        if kwargs.get("type") == "pytorch":
            formatted_attributes["worker_cpu"] = safe_cast(cpu, float)
            formatted_attributes["worker_memory"] = safe_cast(memory, float)
            formatted_attributes["worker_gpu"] = {
                "count": safe_cast(kwargs.get("gpu"), float),
                "mig_device": kwargs.get("mig_device"),
            }

        formatted_attributes.pop('worker_mig_device', None)
        formatted_attributes.pop('mig_device', None)
        formatted_attributes.pop('gpu', None)
        formatted_attributes.pop('labels', None)
        formatted_attributes.pop('taints', None)
        formatted_attributes.pop('shm', None)
        formatted_attributes.pop('huge_pages', None)
        formatted_attributes.pop('gaudi', None)

        attributes_validator(
            {
                **KubeTemplate.available_attributes,
                # These are being saved in server with the parent_ suffix,
                # so they appear without it in available attributes as well.
                "parent_cpu": float,
                "parent_memory": float,
                "parent_labels": str,
                "parent_taints": str,
                "parent_gpu": dict,
                "worker_gpu": dict,
                "type": str,
                "worker_num_executors": int,
                "worker_huge_pages": float,
                "worker_cpu": float,
                "worker_memory": float,
                "worker_labels": str,
                "worker_taints": str,
                "parent_shm": float,
                "worker_shm": float,
                "parent_huge_pages": float,
                "parent_gaudi": float,
            },
            formatted_attributes,
            required_attributes
        )

        # Instance specific validation
        validate_gpu(formatted_attributes["parent_gpu"])
        validate_gpu(formatted_attributes["worker_gpu"])
        validate_template_type(formatted_attributes["type"])

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type='kube_template', attributes=formatted_attributes)
        )

        slug = response.attributes['slug']
        return KubeTemplate(
            context=self._context,
            slug=slug,
            attributes=response.attributes
        )



import sys
import threading
import time
from queue import Empty, Full, Queue

from cnvrgv2.cli.utils.progress_bar_utils import init_progress_bar_for_cli
from cnvrgv2.data.clients.storage_client_factory import storage_client_factory
from cnvrgv2.errors import CnvrgRetryError


class RemoteFilesHandler:
    def __init__(
        self,
        data_owner,
        num_workers=40,
        chunk_size=1000,
        use_cached=False,
        override=False,
        queue_size=5000,
        base_commit_sha1=None,
        commit_sha1=None,
        progress_bar_enabled=False
    ):
        """
        Multithreaded remote file handler base class
        @param data_owner: Cnvrg dataset / project object
        @param num_workers: Number of threads to handle files
        @param chunk_size: File meta chunk size to fetch from the server
        @param override: Override existing files
        @param queue_size: Max number of file meta to put in queue
        """
        # Init data
        self.base_commit_sha1 = base_commit_sha1
        self.commit_sha1 = commit_sha1
        self.progress_bar_enabled = progress_bar_enabled
        self.progress_bar = None

        # Init the storage client
        self.data_owner = data_owner
        self.storage_client = storage_client_factory(refresh_function=data_owner.storage_meta_refresh_function())

        self.use_cached = use_cached
        self.override = override
        self.chunk_size = chunk_size
        self.queue_size = queue_size
        self.progress_queue = Queue(self.queue_size)  # Files deleted / downloaded
        self.handle_queue = Queue(self.queue_size)  # Files need to be deleted / downloaded

        # Create a thread event in order to exit download when needed
        self.task_active = threading.Event()
        self.task_active.set()

        # Create a thread-safe lock
        self.progress_lock = threading.Lock()

        # Create progress thread which tracks the upload progress
        self.errors = None
        self.total_files = sys.maxsize
        self.handled_files = 0
        self.progress_thread = threading.Thread(target=self.task_progress)
        self.progress_thread.start()

        # Create collector thread which fetches file chunks from the server
        self.file_index = 0
        self.collector_thread = threading.Thread(target=self.file_collector)
        self.collector_thread.start()

        # Create downloader threads to parallelize s3 file download
        self.handle_threads = []
        for i in range(num_workers):
            t = threading.Thread(target=self.file_handler)
            t.start()
            self.handle_threads.append(t)

    @property
    def in_progress(self):
        """
        Property used to check if the upload is still in progress
        @return: Boolean
        """
        return self.task_active.is_set()

    def task_progress(self):
        """
        Handles the upload progress and confirming file uploads to the server
        @return: None
        """

        while self.handled_files < self.total_files and self.task_active.is_set():

            try:
                self.progress_queue.get_nowait()

                with self.progress_lock:
                    self.handled_files += 1

            except Empty:
                time.sleep(0.5)

        self.clear()

    def clear(self):
        """
        Clear the threads used to download files
        @return: none
        """
        # Clear download threads
        self.task_active.clear()
        self.collector_thread.join()
        for t in self.handle_threads:
            t.join()

        if self.progress_bar_enabled and self.progress_bar:
            self.progress_bar.finish()

    def file_collector(self):
        """
        The function that handles collecting files metadata from the server
        @return: None
        """
        next_page = 0
        while (
            self.total_files > 0 and
            isinstance(self.file_index, int) and
            self.file_index < self.total_files and
            self.task_active.is_set()
        ):
            # Attempt to retrieve file chunk from the server
            try:
                resp = self._collector_function(page_after=next_page)
                files_to_process = resp["file_dict"]

                # Initialize total_files and progress bar only once
                if self.total_files == sys.maxsize:
                    self.total_files = resp["total_files"]

                if self.progress_bar_enabled and not self.progress_bar:
                    total_files_size = float(resp["total_files_size"])
                    self.progress_bar = init_progress_bar_for_cli("Downloading", total_files_size)

                self.file_index = self.file_index + len(files_to_process)

                # We want to prevent unnecessary loops
                if not resp["next"] or resp["next"] == next_page:
                    return
                else:
                    next_page = resp["next"]

            except Exception as e:
                if self.progress_bar_enabled:
                    print("Could not process files {}".format(self.file_index))
                    print(e)
                # Keep the errors on a variable, and retrieve them back on the main thread
                self.errors = e
                self.task_active.clear()
                return

            # Attempt to put the new files in the download queue, non-blocking in case we want to break iterator loop
            for file in files_to_process:
                while self.task_active.is_set():
                    try:
                        self.handle_queue.put_nowait(file)
                        break
                    except Full:
                        time.sleep(0.5)

    def file_handler(self):
        """
        Function to handle single file
        @return: None
        """
        # Run as long as we have files to download
        while self.task_active.is_set():
            try:
                # Get file non-blocking way, otherwise thread will hang forever
                file = self.handle_queue.get_nowait()
                local_path = "{}/{}".format(self.data_owner.working_dir, file["fullpath"])
                file["local_path"] = local_path

                self._handle_file_function(
                    local_path,
                    object_path=file["object_path"],
                    fullpath=file["fullpath"],
                    cached_commits=file.get("cached_commits", []),
                    type=file.get("type", "files"),
                    progress_bar=self.progress_bar
                )

            except Empty:
                time.sleep(0.5)
            except CnvrgRetryError:
                # If we could not download the file we still count it as processed
                with self.progress_lock:
                    self.handled_files += 1
            except Exception as e:
                # Should not be possible to enter here, safeguard against deadlock
                print("An unhandled exception in downloader thread has occurred:")
                print(e)
                with self.progress_lock:
                    self.handled_files += 1

    def _collector_function(self, page_after=None):
        """
        Base function to collect files metadata from server
        @param page_after: The id of the next file that the iteration of the pagination should start from
        @return: Should return array of files metadata
        """
        pass

    def _handle_file_function(self, local_path, progress_bar=None, **kwargs):
        """
        Base function to handle single file
        @param local_path: File location locally
        @param kwargs: Needs to be fullpath / object_path depends on the class using this base
        @return: None
        """
        pass



import os
from typing import Optional
import boto3
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session
from botocore.config import Config

from cnvrgv2.config import routes, error_messages
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.datasource.datasource_operations import format_credentials, get_refreshable_credentials_function,\
    handle_dir_exist
from cnvrgv2.modules.datasource.datasource_operations_interface import DatasourceOperationsInterface
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.modules.datasource.bucket_downloader import BucketDownloader
from enum import Enum
import logging


# Storage types ENUM
class StorageTypes(Enum):
    S3 = "s3"
    MINIO = "minio"


class Datasource(DatasourceOperationsInterface):
    available_attributes = {
        "slug": str,
        "storage_type": str,
        "bucket_name": str,
        "name": str,
        "admin": str,
        "path": str,
        "endpoint": str,
        "region": str,
        "public": str,
        "description": str,
        "collaborators": str,
        "collaborators_info": str,
        "access_point": str
    }

    def __init__(self, context: Optional[dict] = None, slug: Optional[str] = None,
                 attributes: Optional[dict] = None) -> None:
        """Initialize the Datasource object.

        @param context: The context for the data source.
        @param slug: The slug identifier for the data source.
        @param attributes: Additional attributes for the data source.
        """
        super().__init__()
        self._context = Context(context=context)
        if slug:
            self._context.set_scope(SCOPE.DATASOURCE, slug)
        self.scope = self._context.get_scope(SCOPE.DATASOURCE)
        self._proxy = Proxy(context=self._context)
        self._route = routes.DATASOURCE_BASE.format(self.scope["organization"], self.scope["datasource"])
        self._attributes = attributes or {}
        self.slug = self.scope["datasource"]
        if self.path is None:
            self.path = ""
        elif self.path and not self.path.endswith("/"):
            # if path is not an empty string
            self.path = f"{self.path}/"

        self._client = None

    @property
    def client(self) -> boto3.client:
        """
        Lazy initialization of the S3 client.

        @return: The S3 client.
        """
        if self._client is None:
            self._client = self.get_s3_client()
        return self._client

    def _get_credentials(self) -> dict:
        """Fetch the credentials for the data source.

        @return: The credentials as a dictionary.
        """
        route = urljoin(self._route, "get_credentials")
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.GET
        )
        return response.attributes["config"]

    def get_refreshable_credentials(self) -> dict:
        """Get the refreshable credentials for the data source.

        @return: The refreshable credentials.
        """
        creds = self._get_credentials()
        return format_credentials(creds)

    def get_s3_client(self) -> boto3.client:
        """Get the S3 client with refreshable credentials.

        @return: The S3 client.
        """
        session_credentials = RefreshableCredentials.create_from_metadata(
            metadata=self.get_refreshable_credentials(),
            refresh_using=get_refreshable_credentials_function(self),
            method="get_refreshable_credentials_function"
        )
        session = get_session()
        session._credentials = session_credentials
        session.set_config_variable("region", self.region)
        auto_refresh_session = boto3.Session(botocore_session=session)
        config = Config(max_pool_connections=50)
        return auto_refresh_session.client('s3', endpoint_url=self.endpoint if self.endpoint else None, config=config)

    def add_collaborator(self, collaborator_email: str) -> dict:
        """Add a collaborator to the data source.

        @param collaborator_email: The email of the user to add.
        @return: The response from the API.
        """
        route = urljoin(self._route, "add_collaborator")
        data = {
            "user_email": collaborator_email
        }
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.POST,
            payload=data
        )
        return response

    def remove_collaborator(self, collaborator_email: str) -> dict:
        """Remove a collaborator from the data source.

        @param collaborator_email: The email of the user to remove.
        @return: The response from the API.
        """
        route = urljoin(self._route, "remove_collaborator")
        data = {
            "user_email": collaborator_email
        }
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.POST,
            payload=data
        )
        return response

    def list_objects(self, page_size=100):
        """
        List the objects in the S3 bucket.

        @param page_size: The number of objects to fetch in each page when listing the bucket.
        @type page_size: int
        @return: An iterable paginator of the file names in the bucket path.
        """
        paginator = self.client.get_paginator('list_objects_v2')
        pagination_config = {'PageSize': page_size}
        list_iter = paginator.paginate(Bucket=self.bucket_name, Prefix=self.path, PaginationConfig=pagination_config)

        for page in list_iter:
            relative_paths = []
            for obj in page.get('Contents', []):
                relative_path = obj['Key']
                # Removes datasource.path from the results.
                # (this is because datasource.path represents the "root" path)
                if self.path and relative_path.startswith(self.path):
                    # in the sdk we add a trailing '/' to the path in order for boto3 to list files correctly.
                    # So when removing the datasource_path it will actually remove datasource_path/.
                    # e.g. for key = 'datasource_path/path1/path2', it will return 'path1/path2'
                    # (In s3 folders are saved in the following way : "path1/" "path2/")
                    path_to_remove = self.path
                    if not path_to_remove.endswith("/"):
                        # in this case 'datasource_path/path1/path2', it should return 'path1/path
                        # otherwise it would return '/path1/path2' which would be treated as an absolute path & cause errors
                        path_to_remove = f"{path_to_remove}/"
                    relative_path = relative_path.replace(path_to_remove, "", 1)
                if relative_path:  # don't add empty strings
                    # (this would be for instance if obj = datasource.path, then relative_path would be an empty string
                    relative_paths.append(relative_path)
            yield relative_paths

    def download_file(self, file_path: str, destination_path: str):
        """
        Download a file from the S3 bucket.

        @param file_path: The path of the file in the S3 bucket.
        @type file_path: str
        @param destination_path: The local file path to download the file to.
        @type destination_path: str
        """
        # Key will be datasource.path + file_path
        file_name_normalized = self._normalize_file_path(file_path)
        self.client.download_file(Bucket=self.bucket_name, Key=file_name_normalized,
                                  Filename=destination_path)

        return destination_path

    def get_files_downloader(self, destination_folder: str = None, page_size: int = 100,
                             max_workers: Optional[int] = None) -> BucketDownloader:
        """Get a downloader for downloading multiple files from the S3 bucket.
        @param destination_folder: The file path to download to.
        @param page_size: The number of files to fetch in each page.
        @param max_workers: The maximum number of parallel workers to use for downloading.
        @return: A BucketDownloader instance
        """
        if destination_folder is None:
            destination_folder = os.curdir

        return BucketDownloader(
            datasource=self,
            destination_folder=destination_folder,
            page_size=page_size,
            max_workers=max_workers
        )

    def clone(self, page_size: int = 100, max_workers: Optional[int] = None, skip_if_exists: bool = False,
              force: bool = False) -> None:
        """Clone the contents of the S3 bucket to a local directory.

        @param page_size: The number of files to fetch in each page.
        @param max_workers: The maximum number of parallel workers to use for downloading.
        @param skip_if_exists: If True, skip the cloning if the directory already exists.
        @param force: If True, forcefully delete the existing directory.
        """
        # Destination folder for clone: /datasource_slug
        destination_folder = self.slug

        # Handle existing datasource_slug folder:
        # 1. By default, if the folder already exists, raise an error and return.
        # 2. Force: delete the existing folder and download all. (This is useful when we want to make sure we
        # are aligned with the latest version of the datasource.)
        # 3. skip if exists: if the folder already exists, return without raising an error. (This is useful when running
        # jobs, if the user knows no changes have been made since last clone)
        if os.path.isdir(destination_folder):
            if skip_if_exists:
                if force:
                    raise CnvrgArgumentsError(error_messages.DATASOURCE_BAD_FORCE_ARGUMENTS)
                return
            else:
                handle_dir_exist(force, destination_folder)

        downloader = self.get_files_downloader(page_size=page_size,
                                               max_workers=max_workers,
                                               destination_folder=destination_folder)

        for page in downloader.page_iterator:
            downloader.download_objects(page)

    def upload_file(self, file_path: str, destination_path: Optional[str] = None) -> None:
        """Upload a file to the S3 bucket.
         @param file_path: The local file to upload.
         @param destination_path: The destination file path in the bucket.
         @return response: S3 response
        """
        if not destination_path:
            destination_path = file_path
        file_name_normalized = self._normalize_file_path(destination_path)
        response = self.client.upload_file(Bucket=self.bucket_name, Filename=file_path, Key=file_name_normalized)
        return response

    def remove_file(self, file_path: str) -> None:
        """Remove a file from the S3 bucket.
        @param file_path: The path of the file to remove.
        """
        file = self._normalize_file_path(file_path)
        response = self.client.delete_objects(Bucket=self.bucket_name, Delete={'Objects': [{'Key': file}]})
        if 'Errors' in response:
            error_message = ", ".join([error['Message'] for error in response['Errors']])
            raise Exception(f"Failed to delete file '{file_path}']: {error_message}")

    def _normalize_file_path(self, file_path: str) -> str:
        """Normalize the file name to include the path.
        @param file_path: The file name to normalize.
        @return: The file name normalized with datasource path.
        """
        if not self.path:
            return file_path
        else:
            relative_path = file_path.lstrip('/')  # if file_path is an absolute path, join will ignore self.path
            return os.path.join(self.path, relative_path)

    def delete(self) -> None:
        """Delete the data source."""
        logging.info("Deleting the datasource may affect running jobs that use it")
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)



import unittest
from unittest.mock import patch
from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgArgumentsError
import pytest


class TestDatasourceClone:
    page_size = 50
    max_workers = 4

    @patch('cnvrgv2.modules.datasource.datasource.handle_dir_exist')
    @patch('os.path.isdir')
    def test_clone(self, mock_isdir,
                   mock_handle_dir_exist, datasource, mock_datasource_client, mock_datasource_files_downloader):
        # Clone with default params: skip_if_exists=False, force=False when datasource has not already been cloned
        # Datasource folder does not exist
        mock_isdir.return_value = False

        # Call the clone method

        datasource.clone(page_size=self.page_size, max_workers=self.max_workers, skip_if_exists=False, force=False)

        # Datasource should be cloned to datasource.slug folder
        destination_folder = datasource.slug
        mock_isdir.assert_called_once_with(destination_folder)
        mock_handle_dir_exist.assert_not_called()

        datasource.get_files_downloader.assert_called_once_with(
            page_size=self.page_size,
            max_workers=self.max_workers,
            destination_folder=destination_folder
        )
        # Ensure the downloader download_objects method is called for each page
        expected_calls = [unittest.mock.call(page) for page in datasource.get_files_downloader.page_iterator]
        datasource.get_files_downloader.download_objects.assert_has_calls(expected_calls, any_order=False)

    @patch('cnvrgv2.modules.datasource.datasource.handle_dir_exist')
    @patch('os.path.isdir')
    def test_clone_skip_if_exists(self, mock_isdir, mock_handle_dir_exist, datasource, mock_datasource_client,
                                  mock_datasource_files_downloader):
        # Clone with params: skip_if_exists=True, force=False when datasource has already been cloned
        # Datasource folder exists
        mock_isdir.return_value = True

        # Call the clone method with skip_if_exists=True
        datasource.clone(page_size=self.page_size, max_workers=self.max_workers, skip_if_exists=True, force=False)

        mock_isdir.assert_called_once_with(datasource.slug)
        mock_handle_dir_exist.assert_not_called()
        datasource.get_files_downloader.assert_not_called()

    @patch('cnvrgv2.modules.datasource.datasource.handle_dir_exist')
    @patch('os.path.isdir')
    def test_clone_force_delete(self, mock_isdir, mock_handle_dir_exist, datasource, mock_datasource_client,
                                mock_datasource_files_downloader):
        # Clone with params: skip_if_exists=False, force=True when datasource has already been cloned
        # Datasource folder exists
        mock_isdir.return_value = True

        # Call the clone method with force=True
        datasource.clone(page_size=self.page_size, max_workers=self.max_workers, skip_if_exists=False, force=True)

        # Assertions
        mock_isdir.assert_called_once_with(datasource.slug)
        mock_handle_dir_exist.assert_called_once_with(True, datasource.slug)

        datasource.get_files_downloader.assert_called_once_with(
            page_size=self.page_size,
            max_workers=self.max_workers,
            destination_folder=datasource.slug,
        )
        # Ensure the downloader download_objects method is called for each page
        expected_calls = [unittest.mock.call(page) for page in datasource.get_files_downloader.page_iterator]
        datasource.get_files_downloader.download_objects.assert_has_calls(expected_calls, any_order=False)

    @patch('cnvrgv2.modules.datasource.datasource.handle_dir_exist')
    @patch('os.path.isdir')
    def test_clone_args_error(self, mock_isdir, mock_handle_dir_exist, datasource, mock_datasource_client):
        # Clone with params: skip_if_exists=True, force=True - both cannot be set to True
        mock_isdir.return_value = True

        # Call the clone method with skip_if_exists=True and force=True (should raise error)
        with pytest.raises(CnvrgArgumentsError) as exception_info:
            datasource.clone(skip_if_exists=True, force=True)

        # Check if the exception message matches the expected error message
        assert error_messages.DATASOURCE_BAD_FORCE_ARGUMENTS in str(exception_info.value)

        # Assert handle_dir_exist was not called because of the error
        mock_handle_dir_exist.assert_not_called()



from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.libhub.libraries.library_version import LibraryVersion


class LibraryVersionsClient:
    def __init__(self, library):

        self._context = Context(context=library._context)

        scope = self._context.get_scope(SCOPE.LIBRARY)
        self.library = library
        self._route = routes.LIBRARY_VERSIONS_BASE.format(scope["organization"], scope["library"])

    def list(self, sort="-id"):
        """
        List all projects in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields project objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=LibraryVersion,
            sort=sort
        )

    def get(self, version):
        """
        Retrieves a project by the given slug
        @param slug: The slug of the requested project
        @return: Project object
        """
        if not version or not isinstance(version, str):
            raise CnvrgArgumentsError(error_messages.PROJECT_GET_FAULTY_SLUG)

        return LibraryVersion(context=self._context, version=version)

    @property
    def latest(self):
        return self.get(version="latest")



import click
from cnvrgv2.modules.workflows import NotebookType
from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command


@click.group(name='workspace')
def workspace_group():
    pass


@workspace_group.command()
@click.option('-t', '--title', default=None, help=messages.WORKSPACE_HELP_TITLE)
@click.option('-tm', '--templates', default=None, help=messages.WORKSPACE_HELP_TEMPLATES)
@click.option('-c', '--commit', default=None, help=messages.WORKSPACE_HELP_COMMIT)
@click.option('-nt', '--notebook_type', default=NotebookType.JUPYTER_LAB, help=messages.WORKSPACE_HELP_NOTEBOOK_TYPE)
@click.option('-d', '--datasets', default=None, help=messages.WORKSPACE_HELP_DATASETS)
@click.option('-ds', '--datasources', default=[], help=messages.WORKSPACE_HELP_DATASOURCE)
@click.option('-v', '--volume', default=None, help=messages.WORKSPACE_HELP_VOLUME)
@click.option('-i', '--image', default=None, help=messages.WORKSPACE_HELP_IMAGE)
@click.option('-gb', '--git-branch', default=None, help=messages.WORKSPACE_HELP_GIT_BRANCH)
@click.option('-gc', '--git-commit', default=None, help=messages.WORKSPACE_HELP_GIT_COMMIT)
@click.option('-lf', '--local_folders', default=None, help=messages.WORKSPACE_HELP_LOCAL_FOLDERS)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def create(cnvrg, project, logger, title, commit, templates, notebook_type, datasets, datasources, volume,
           image, git_branch, git_commit, local_folders, project_slug):
    """
      Create a workspace
    """
    dataset_objects = None
    volume_object = None
    kwargs = {}
    templates_list = templates.split(",") if templates else None

    if datasets:
        dataset_names = datasets.split(",")
        dataset_objects = [cnvrg.datasets.get(ds_name) for ds_name in dataset_names]

    if volume:
        volume_object = project.volumes.get(volume)

    if image:
        image_name, image_tag = image.split(":")
        kwargs["image"] = cnvrg.images.get(name=image_name, tag=image_tag)

    if git_branch:
        kwargs["git_branch"] = git_branch

    if git_commit:
        kwargs["git_commit"] = git_commit

    workspace = project.workspaces.create(
        title=title,
        templates=templates_list,
        notebook_type=notebook_type,
        datasets=dataset_objects,
        volume=volume_object,
        commit=commit,
        local_folders=local_folders,
        datasource_slugs=datasources,
        **kwargs
    )

    success_message = messages.WORKSPACE_CREATE_SUCCESS.format(workspace.title, workspace.full_href)
    logger.log_and_echo(success_message)


@workspace_group.command()
@click.option('-s', '--slug', default=None, help=messages.WORKSPACE_SLUG)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def start(workspace, logger, slug, project_slug):
    """
      starts the current workspace
    """
    if workspace is None:
        logger.log_and_echo(messages.WORKSPACE_DOES_NOT_EXIST, error=True)

    workspace.start()
    logger.log_and_echo('Workflow {} started!'.format(slug))


@workspace_group.command()
@click.option('-s', '--slug', default=None, help=messages.WORKSPACE_SLUG)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def sync_remote(workspace, logger, slug, project_slug):
    """
      sync the current workspace
    """
    if workspace is None:
        logger.log_and_echo(messages.WORKSPACE_DOES_NOT_EXIST, error=True)

    workspace.sync_remote()
    logger.log_and_echo('Workflow {} sync has been sent'.format(slug))


@workspace_group.command()
@click.option('-s', '--slug', default=None, help=messages.WORKSPACE_SLUG)
@click.option('-sy', '--sync', default=True, help=messages.WORKSPACE_HELP_SYNC_REMOTE)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def stop(workspace, logger, slug, sync, project_slug):
    """
      stop the current workspace
    """
    if workspace is None:
        logger.log_and_echo(messages.WORKSPACE_DOES_NOT_EXIST, error=True)

    workspace.stop(sync=sync)
    logger.log_and_echo('Workspace {} stopped!'.format(slug))


@workspace_group.command()
@click.option('-s', '--slug', default=None, help=messages.WORKSPACE_SLUG)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def start_tensorboard(workspace, logger, slug, project_slug):
    """
      starts tensorboard associated with the current workspace
    """
    if workspace is None:
        logger.log_and_echo(messages.WORKSPACE_DOES_NOT_EXIST, error=True)

    workspace.start_tensorboard()
    logger.log_and_echo('tensorboard started!')


@workspace_group.command()
@click.option('-s', '--slug', default=None, help=messages.WORKSPACE_SLUG)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def stop_tensorboard(workspace, logger, slug, project_slug):
    """
      stops tensorboard associated with the current workspace
    """
    if workspace is None:
        logger.log_and_echo(messages.WORKSPACE_DOES_NOT_EXIST, error=True)

    workspace.stop_tensorboard()
    logger.log_and_echo('tensorboard session stopped')



import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Optional, List

from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.datasource.datasource_operations_interface import DatasourceOperationsInterface


class BucketDownloader:
    def __init__(self, datasource: DatasourceOperationsInterface, destination_folder: str, page_size: int = 100,
                 max_workers: Optional[int] = None) -> None:
        self.datasource = datasource
        self.destination_folder = os.path.abspath(destination_folder)
        self.max_workers = max_workers if max_workers is not None else get_max_workers()
        self.page_size = page_size
        self.page_iterator = self.datasource.list_objects(page_size=page_size)
        os.makedirs(self.destination_folder, exist_ok=True)

    def download_objects(self, objects: List[dict]) -> None:
        """Download multiple objects from S3 using a thread pool."""
        try:
            with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                # Submit all download tasks and collect Future objects
                # Submit datasource.download_file foreach file_name in the list
                future_to_file = {
                    executor.submit(self.pre_download, file_path): file_path
                    for file_path in objects
                }

                # Iterate over the Future objects as they complete
                for future in as_completed(future_to_file):
                    file_path = future_to_file[future]
                    try:
                        # This will raise an exception if the task raised one
                        future.result()
                    except Exception as error:
                        raise CnvrgError(f"An error occurred while processing {file_path}: {error}")
                        # Cancel remaining futures
                        for pending_future in future_to_file:
                            if not pending_future.done():
                                pending_future.cancel()
                        break  # Exit the loop after canceling the remaining tasks
        except Exception as exc:
            raise CnvrgError(f"Clone error: {exc}")

    def pre_download(self, file_path):
        # destination_path is: self.destination_folder + file_name
        # (where destination_folder should be the datasource slug)
        destination = self._calc_destination_file_path(file_name=file_path)
        directory = os.path.dirname(destination)
        # Prepare destination folder if it doesn't exist: it might be path1/path2/file_name.txt.
        # In this cas we would need to create path1/ and path1/path2/
        if directory:
            os.makedirs(directory, exist_ok=True)
        if not file_path.endswith('/'):
            self.datasource.download_file(file_path=file_path, destination_path=destination)

    def _calc_destination_file_path(self, file_name):
        relative_path = file_name.lstrip('/')  # if file_name is an absolute path, join will ignore self.destination_folder
        return os.path.join(self.destination_folder, relative_path)


def get_max_workers() -> int:
    """
    For IO-bound tasks ( e.g., downloading files from S3),
    a commonly used formula is: max_workers = 2 × number_of_cores + 1
    @return: num of recommended workers
    """
    return 2 * os.cpu_count() + 1



import pytest

from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgError
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.registries import RegistriesClient, Registry, RegistryTypes


class MockRegistry:
    def __init__(self, list=False):
        self.slug = "test-slug"
        self.url = "doker.io"
        self.type = RegistryTypes.OTHER
        self.title = "my-registry"
        self.username = "username"

        self.data = {
            "id": "17",
            "type": "registry",
            "attributes": {
                "slug": self.slug,
                "url": self.url,
                "title": self.title,
                "private": True,
                "username": self.username,
                "registry_type": self.type
            }
        }
        if list:
            self.data = [self.data]

    def response(self, *args, **kwargs):
        return JAF(response={"data": self.data})


class TestRegistriesClient:
    def test_create_registry_with_valid_arguments(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockRegistry()
        mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        client = RegistriesClient(empty_cnvrg)
        image = client.create(title=mock_resp.title, url=mock_resp.url, type=mock_resp.type)
        assert isinstance(image, Registry)
        assert image.slug == mock_resp.slug

    def test_create_registry_with_invalid_arguments(self, empty_context, empty_cnvrg):
        url = "docker.io"
        type = RegistryTypes.OTHER
        title = "test-image"

        client = RegistriesClient(empty_cnvrg)
        with pytest.raises(CnvrgError):
            client.create(title=1, url=url, type=type)

        with pytest.raises(CnvrgError):
            client.create(title=title, url=1, type=type)

        with pytest.raises(CnvrgError):
            client.create(title=title, url="badurl", type=type)

        with pytest.raises(CnvrgError):
            client.create(title=title, url=url, type=1)

        with pytest.raises(CnvrgError):
            client.create(title=title, url=url, type="fake")

        with pytest.raises(CnvrgError):
            client.create(title=title, url=url, type=type, username=1)

    def test_list_registries(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockRegistry(list=True)
        mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        client = RegistriesClient(empty_cnvrg)

        registries = client.list()
        registries = [registry for registry in registries]

        assert len(registries) == 1
        assert isinstance(registries[0], Registry)
        assert registries[0].slug == mock_resp.slug

    def test_get_registry(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockRegistry()
        mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        client = RegistriesClient(empty_cnvrg)
        registry = client.get(mock_resp.slug)

        assert isinstance(registry, Registry)
        assert registry.slug == mock_resp.slug
        assert registry.title == mock_resp.title

    def test_delete_image(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockRegistry()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        client = RegistriesClient(empty_cnvrg)
        client.delete(mock_resp.slug)

        assert mocked_call.call_count == 1



import mimetypes
import os
import threading
import time
from queue import Empty, Full, Queue

from cnvrgv2.cli.utils.progress_bar_utils import init_progress_bar_for_cli
from cnvrgv2.data.clients.storage_client_factory import storage_client_factory
from cnvrgv2.utils.converters import convert_bytes
from cnvrgv2.utils.storage_utils import append_trailing_slash, get_file_sha1


class LocalFilesHandler:
    def __init__(
        self,
        data_owner,
        paths,
        metadata,
        override=False,
        num_workers=40,
        chunk_size=1000,
        queue_size=5000,
        progress_bar_enabled=False,
        dir_path=""
    ):
        """
        Multithreaded local file handler base class
        @param data_owner: Cnvrg dataset / project object
        @param paths: [Generator] that lists paths to upload
        @param metadata: [Dict] The metadata on the uploaded files (total files, file size..)
        @param override: Boolean stating whether or not we should re-upload even if the file already exists
        @param num_workers: Number of threads to use for concurrent file handling
        @param chunk_size: File meta chunk size to fetch from the server
        @param queue_size: Max number of file meta to put in queue
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        @param dir_path: The destination directory path where the folders and files will be uploaded in cnvrg
        """
        # Init the storage client
        self.data_owner = data_owner
        self.storage_client = storage_client_factory(refresh_function=data_owner.storage_meta_refresh_function())

        # Init helper vars
        self.override = override
        self.queue_size = queue_size
        self.chunk_size = chunk_size
        self.path_generator = paths
        self.progress_bar_enabled = progress_bar_enabled
        self.errors = []
        self.progress_bar = None
        self.dir_path = dir_path
        if progress_bar_enabled:
            if metadata["total_size"] > 0:
                self.progress_bar = init_progress_bar_for_cli("Uploading", metadata["total_size"])
            else:
                self.progress_bar_enabled = False
                print("Total size is 0, skipping progress bar")

        # Init file queues
        self.progress_queue = Queue(self.queue_size)
        self.handle_queue = Queue(self.queue_size)

        # Create a thread event in order to exit handling when needed
        self.handling_active = threading.Event()
        self.handling_active.set()

        # Create a thread-safe lock
        self.progress_lock = threading.Lock()

        # Create collector thread which sends file chunks to the server
        self.collector_thread = threading.Thread(target=self.file_collector)
        self.collector_thread.start()

        # Create progress thread which tracks the upload progress
        self.total_files = metadata["total_files"]
        self.handled_files = 0
        self.progress_thread = threading.Thread(target=self.task_progress)
        self.progress_thread.start()

        # Create bucket_downloader.py threads to parallelize file handling
        self.handler_threads = []
        for i in range(num_workers):
            t = threading.Thread(target=self.file_handler)
            t.start()
            self.handler_threads.append(t)

    def clear(self):
        """
        Clear the threads used to upload files
        @return: none
        """
        # Clear download threads
        self.handling_active.clear()
        self.collector_thread.join()

        try:
            for t in self.handler_threads:
                t.join()

            if self.progress_bar_enabled:
                self.progress_bar.finish()
        except Exception:
            pass

    @property
    def in_progress(self):
        """
        Property used to check if the upload is still in progress
        @return: Boolean
        """
        return self.handling_active.is_set()

    def file_collector(self):
        """
        The function that handles collecting files metadata from the server
        @return: None
        """

        files_chunk = {}
        data_owner_root = append_trailing_slash(self.data_owner._config.root) if self.data_owner._config.root else ''

        for fullpath in self.path_generator:
            # Normalize fullpath to be relative to the working directory
            fullpath = fullpath.replace(data_owner_root, "")
            is_file = os.path.isfile(fullpath)
            is_directory = os.path.isdir(fullpath)
            relative_path = fullpath
            if self.dir_path:
                relative_path = relative_path.replace(os.getcwd(), "")
                relative_path = os.path.join(self.dir_path, relative_path.lstrip('/'))

            if is_file:
                content_type = mimetypes.guess_type(fullpath, strict=True)[0] or "plain/text"
                files_chunk[relative_path] = {
                    "sha1": get_file_sha1(fullpath),
                    "file_name": os.path.basename(fullpath),
                    "file_size": os.path.getsize(fullpath),
                    "local_path": fullpath,
                    "content_type": content_type,
                    # TODO: Might be an unnecessary duplicate of line 116. Check and remove the replace if not needed
                    "relative_path": relative_path.replace(append_trailing_slash(self.data_owner.working_dir), ''),
                }
            elif is_directory:
                # For directories we don't need any metadata
                files_chunk[relative_path] = None

            if len(files_chunk) % self.chunk_size == 0:
                self._handle_collector_files(files_chunk)
                files_chunk.clear()

        # Handle remaining chunk
        self._handle_collector_files(files_chunk)

    def _handle_collector_files(self, files_chunk):
        """
        Helper function that handles calling the server and pushes the result to the worker threads
        @param files_chunk: Chunk of files to process
        @return: None
        """
        try:
            files_to_upload = self._file_collector_function(files_chunk)
        except Exception as e:
            self.errors.append(e)
            if self.progress_bar_enabled:
                print("Could not process files")
                print(e)

            self.handling_active.clear()
            return

        self._handle_collector_progress(files_chunk, files_to_upload)

        # Attempt to put the new files in the upload queue, non-blocking in case we want to stop the upload
        for file in files_to_upload:
            while self.handling_active.is_set():
                try:
                    self.handle_queue.put_nowait(file)
                    break
                except Full:
                    time.sleep(0.5)

    def _handle_collector_progress(self, chunk, files_to_upload):
        """
        Handles upload progress in case the some of the files we were trying to upload already exist in the server and
        do not need to be re-uploaded again.
        @param chunk: [Dict] The file metadata we sent to the server
        @param files_to_upload: [List] The list of files that needed to get uploaded
        @return: None
        """
        with self.progress_lock:
            self.handled_files += len(chunk) - len(files_to_upload)

            # Show progress for files already in server
            if self.progress_bar_enabled:
                # Calculate total file size for files (minus folders) to upload:
                chunk_file_size = sum([chunk[fullpath]["file_size"] for fullpath in chunk if chunk[fullpath]])
                # Calculate total file size of the files needed to upload:
                upload_file_size = sum([file["file_size"] for file in files_to_upload])
                # Calculate the size of the file we are NOT going to upload and add them to the progress
                converted_bytes, unit = convert_bytes(chunk_file_size - upload_file_size, self.progress_bar.unit)
                self.progress_bar.next(converted_bytes)

    def _normalize_file_name(self, file_name):
        if file_name.startswith('/'):
            return file_name[1:]
        return file_name

    def file_handler(self):
        """
        Handles uploading files to the relevant object storage
        @return: None
        """
        # Run as long as we have files to upload
        file = None

        while self.handling_active.is_set():
            try:
                # Get file non-blocking way, otherwise thread will hang forever
                file = self.handle_queue.get_nowait()
                self._file_handler_function(file["local_path"], file.get("object_path"), progress_bar=self.progress_bar)
                self.handle_queue.task_done()

                self.progress_queue.put(file)
            except Empty:
                time.sleep(0.5)
            except Exception as e:
                self.errors.append(e)
                if self.progress_bar_enabled:
                    print("could not upload file {}".format(file["local_path"]))
                    print(e)
                with self.progress_lock:
                    self.handled_files += 1

    def task_progress(self):
        """
        Handles the upload progress and confirming file uploads to the server
        @return: None
        """
        uploaded_ids = []

        while self.handled_files < self.total_files and self.handling_active.is_set():
            try:
                file = self.progress_queue.get_nowait()

                with self.progress_lock:
                    self.handled_files += 1

                uploaded_ids.append(file["id"])

                if len(uploaded_ids) >= self.chunk_size or self.handled_files >= self.total_files:
                    try:
                        self._handle_file_progress_function(uploaded_ids)
                    finally:
                        uploaded_ids = []
            except Empty:
                time.sleep(0.5)

        if len(uploaded_ids) > 0:
            self._handle_file_progress_function(uploaded_ids)

        self.clear()

    def _handle_file_progress_function(self, uploaded_ids):
        """
        Base function to progress the task
        @param uploaded_ids: Uploaded file ids
        @return: None
        """
        pass

    def _file_handler_function(self, local_path, object_path, progress_bar=None):
        """
        Base function to handle single file
        @param local_path: File location locally
        @param object_path: File location in bucket
        @return: None
        """
        pass

    def _file_collector_function(self, files_metadata):
        """
        Base function to collect files metadata from server
        @param files_metadata: Local files metadata to sent to the server for getting files to handle
        @return: Should return array of files metadata
        """
        pass



import time
from functools import wraps

from cnvrgv2.errors import CnvrgRetryError


def retry(tries=10, delay=5, backoff=2, log_error=False):
    """
    Will retry decorated function with exponential backoff
    @param tries: Number of retries
    @param delay: Initial delay between attempts
    @param backoff: exponential backoff factor
    @param log_error: Whether or not print the raised error
    @return: decorator for retrying a function
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempt = 0
            backoff_delay = delay
            while attempt < tries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if log_error:
                        print(e)
                    time.sleep(backoff_delay)
                    attempt += 1
                    backoff_delay *= backoff
            raise CnvrgRetryError("Number of retries exceeded")
        return wrapper
    return decorator



import os

import requests

from cnvrgv2.config import Config, CONFIG_VERSION
from cnvrgv2.config import error_messages, routes
from cnvrgv2.config.config import SSO_VERSION
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgError, CnvrgHttpError, CnvrgLoginError
from cnvrgv2.modules.users.users_client import UsersClient
from cnvrgv2.proxy import HTTP, Proxy


class SCOPE:
    PROJECT = {
        "key": "project",
        "dependencies": ["organization", "user"]
    }
    IMAGE = {
        "key": "image",
        "dependencies": ["organization", "user"]
    }
    REGISTRY = {
        "key": "registry",
        "dependencies": ["organization", "user"]
    }
    LABEL = {
        "key": "label",
        "dependencies": ["organization", "user"]
    }
    VOLUME = {
        "key": "volume",
        "dependencies": ["project", "organization", "user"]
    }
    DATASET = {
        "key": "dataset",
        "dependencies": ["organization", "user"]
    }
    MEMBER = {
        "key": "member",
        "dependencies": ["organization"]
    }
    LIBRARY = {
        "key": "library",
        "dependencies": ["organization", "user"]
    }
    LIBRARY_VERSION = {
        "key": "library_version",
        "dependencies": ["library", "organization", "user"]
    }
    COMMIT = {
        "key": "commit",
        "dependencies": ["dataset", "organization", "user"]
    }
    QUERY = {
        "key": "query",
        "dependencies": ["dataset", "organization", "user"]
    }
    RESOURCE = {
        "key": "resource",
        "dependencies": ["organization", "user"]
    }
    TEMPLATE = {
        "key": "template",
        "dependencies": ["resource", "organization", "user"]
    }
    EXPERIMENT = {
        "key": "experiment",
        "dependencies": ["project", "organization", "user"]
    }
    WORKSPACE = {
        "key": "notebooksession",
        "dependencies": ["project", "organization", "user"]
    }
    WEBAPP = {
        "key": "webapp",
        "dependencies": ["project", "organization", "user"]
    }
    ENDPOINT = {
        "key": "endpoint",
        "dependencies": ["project", "organization", "user"]
    }
    FLOW = {
        "key": "flow",
        "dependencies": ["project", "organization", "user"]
    }
    FLOW_VERSION = {
        "key": "flow-version",
        "dependencies": ["flow", "project", "organization", "user"]
    }
    ORGANIZATION = {
        "key": "organization",
        "dependencies": ["user"]
    }
    DATASOURCE = {
        "key": "datasource",
        "dependencies": ["organization", "user"]
    }

    @classmethod
    def scopes(cls):
        """
        Calculates all of the scopes from the class attributes
        @return: scope list
        """
        scopes = []
        for attr in dir(cls):
            if not callable(getattr(cls, attr)) and not attr.startswith("__"):
                scopes.append(getattr(cls, attr)["key"])
        return scopes


class Context:
    def __init__(
            self, context=None, domain=None, user=None, password=None, organization=None, token=None,
            sso_version=SSO_VERSION.CAPI
            ):
        # Set the credentials variables:
        self.token = None
        self.domain = None
        self.user = None
        self.sso_version = sso_version

        # Set scope variables:
        for scope in SCOPE.scopes():
            setattr(self, scope, None)

        # If a context is passed, perform a deep copy and return
        if context:
            self._copy_context(context=context, organization=organization)
            return

        # Cannot pass username/password without a domain+username+password or domain + user + token
        if (user or password) and not all([domain, user, password]) and not all([domain, user, token]):
            raise CnvrgArgumentsError(error_messages.CONTEXT_BAD_ARGUMENTS)

        # If a domain is passed, init a blank Cnvrg client
        if domain:
            self.domain = domain

            # Attempt to authenticate using the credentials if they were passed
            if user:
                self._authenticate(domain, user, password, token)
        else:
            self._load_credentials()

        # Override organization if it was explicitly passed
        if organization:
            self.organization = organization

        self.ensure_organization_exist(self.organization)

    def get_scope(self, scope):
        """
        Checks if the current context contains the relevant scope
        @param scope: a SCOPE object constant
        @raise CnvrgError: if the context does not contain the required scope dependencies
        @return: scope dictionary
        """

        dependencies_dict = self._check_dependencies(scope)
        cur_scope = getattr(self, scope["key"], None)
        if cur_scope is None:
            error_msg = error_messages.CONTEXT_SCOPE_BAD_SCOPE.format(scope["key"])
            raise CnvrgError(error_msg)

        return {scope["key"]: cur_scope, **dependencies_dict}

    def set_scope(self, scope, slug):
        """
        Checks if the current context contains the relevant scope and sets new scope
        @param scope: a SCOPE object constant
        @param slug: scope object slug
        @raise CnvrgError: if the context does not contain the required scope dependencies
        @return: slug/None
        """
        if scope == SCOPE.ORGANIZATION:
            self.ensure_organization_exist(slug)

        self._check_dependencies(scope)
        setattr(self, scope["key"], slug)

    def save(self):
        """
        Creates a .cnvrg/config file to reuse authentication.
        Config save will only work in organization context, otherwise an error will be raised
        @raise CnvrgError: Tried to save without having at least organization context
        @return: Boolean depending if context save was successful
        """
        config = Config()
        # Will throw a CnvrgError if organization context is not present
        config_data = self._generate_context_dict()
        config.update(**config_data)

    def _copy_context(self, context, organization=None):
        """
        Performs a deep copy of the supplied context into the current one
        @param context: The source context
        @param organization: Organization name to override
        @return: None
        """
        self.token = context.token
        self.sso_version = context.sso_version
        self.domain = context.domain
        self.user = context.user

        for scope in SCOPE.scopes():
            target_attr = getattr(context, scope, None)
            setattr(self, scope, target_attr)

        if organization:
            self.organization = organization

    def _check_dependencies(self, scope):
        """
        This function checks that the requested scope has all of its dependencies initialized
        @param scope: The scope const we want to check
        @return: dict of scopes and their slugs
        """
        # Build scope object
        scope_dict = {}
        for dependency in scope["dependencies"]:
            dependency_slug = getattr(self, dependency, None)
            if dependency_slug is None:
                error_msg = error_messages.CONTEXT_SCOPE_BAD_DEPENDENCIES.format(scope["key"], dependency)
                raise CnvrgError(error_msg)
            else:
                scope_dict[dependency] = dependency_slug

        return scope_dict

    def _authenticate(self, domain, user, password, user_token=None):
        """
        Performs authentication against Cnvrg and retrieves the auth token
        @param domain: Cnvrg app domain
        @param user: Email with which the user was registered
        @param password: Password with which the user was registered
        @param user_token: user authentication token instead of password
        @return: None
        """
        auth = UsersClient(domain=domain, token=user_token)
        self.sso_version = auth.sso_version

        if password is None:
            password = ""

        # Will raise an exception if login did not succeed
        token, organization, sso_enabled = auth.login(user=user, password=password, token=user_token)

        if user_token is not None and sso_enabled:
            self.token = user_token
        else:
            self.token = token
        self.user = user
        self.organization = organization

        # We don't allow users without an organization to use the context
        if organization is None:
            raise CnvrgError(error_messages.USER_NO_ORGANIZATION)

    def _load_credentials(self):
        """
        Attempts to login using either environment variables or the global Cnvrg config file
        @raise: CnvrgError if cannot login using either method
        @return: None
        """
        logged_in = self._load_from_env() or self._load_from_config()
        if not logged_in:
            raise CnvrgLoginError(error_messages.CONTEXT_BAD_ENV_VARIABLES)

    def _load_from_env(self):
        """
        Attempts to find credentials in environment variables
        @return: Boolean
        """
        # user token is the API token
        # In cnvrg jobs this can be the API token or the SSO token
        user_token = os.environ.get("CNVRG_TOKEN", None)
        jwt_token = os.environ.get("CNVRG_JWT_TOKEN", None)
        provided_token = user_token or jwt_token
        domain = os.environ.get("CNVRG_URL")
        user = os.environ.get("CNVRG_USER")
        organization = os.environ.get("CNVRG_ORGANIZATION")

        if not all([provided_token, domain, user, organization]):
            return False

        self.domain = domain
        self.user = user
        self.organization = organization
        self.sso_version = os.environ.get("CNVRG_SSO_VERSION", SSO_VERSION.CAPI)

        # Pod-specific env vars
        project = os.environ.get("CNVRG_PROJECT")
        dataset = os.environ.get("CNVRG_DATASET")

        if project:
            self.project = project
        if dataset:
            self.dataset = dataset

        job_slug = os.environ.get("CNVRG_JOB_ID")
        job_type = os.environ.get("CNVRG_JOB_TYPE")

        if job_type is not None:
            setattr(self, job_type.lower(), job_slug)

        # In case user is logging in using environment variables
        # This might be the user's simple token which should be exchanged for a JWT token
        if user_token and jwt_token is None:
            self._authenticate(self.domain, self.user, None, user_token)
        else:
            self.token = jwt_token
        return True

    def _load_from_config(self):
        """
        Attempts to find credentials in local/global config file
        @return: Boolean
        """
        config = Config()
        if not config:
            return False

        token, domain, user = config.get_credential_variables()

        # Context variables
        organization = config.organization

        if not all([token, domain, user, organization]):
            return False

        self.token = token
        self.sso_version = config.sso_version
        self.domain = domain
        self.user = user
        self.organization = organization

        if config.project_slug:
            self.project = config.project_slug
        if config.dataset_slug:
            self.dataset = config.dataset_slug

        return True

    def _generate_context_dict(self):
        """
        Generates a context dict to save in a local config file
        @return: dict
        """
        if not all([self.token, self.domain, self.user, self.organization]):
            raise CnvrgError(error_messages.CONTEXT_CANT_SAVE)

        context = {
            "user": self.user,
            "token": self.token,
            "domain": self.domain,
            "version": CONFIG_VERSION,
            "organization": self.organization,
            "check_certificate": False
        }

        if self.project:
            context["project_slug"] = self.project

        if self.dataset:
            context["dataset_slug"] = self.dataset

        return context

    def ensure_organization_exist(self, name):
        try:
            proxy = Proxy(domain=self.domain, token=self.token, sso_version=self.sso_version)
            organization_route = routes.ORGANIZATION_BASE.format(name)

            proxy.call_api(route=organization_route, http_method=HTTP.GET)

        except CnvrgHttpError as e:
            if e.status_code == requests.codes.not_found:
                raise CnvrgArgumentsError(error_messages.ORGANIZATION_DOESNT_EXIST)
            else:
                # Don't suppress unexpected exceptions
                raise e

    def get_env_variables(self):
        """
        Attempts to find credentials in environment variables
        @return: token, domain, user
        """
        # Credential variables
        token = self.token
        domain = self.domain
        user = self.user
        return token, domain, user



from cnvrgv2.modules.organization.organization_settings import OrganizationSettings
from cnvrgv2.proxy import Proxy


class TestOrganizationSettings:

    def test_save_organization_settings(self, mocker, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        organization_settings = OrganizationSettings(organization=empty_cnvrg)

        mocked_save = mocker.patch.object(Proxy, 'call_api')
        organization_settings.save()
        assert mocked_save.call_count == 1



from cnvrgv2.config import routes
from cnvrgv2.errors import CnvrgHttpError
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.utils.validators import attributes_validator

DISABLED_PROPERTY_INT = -1
DISABLED_PROPERTY_STRING = ""
STRING_PROPERTIES = ("slack_webhook_url", "custom_pypi_url")


class OrganizationSettings(DynamicAttributes):
    available_attributes = {
        "default_computes": list,
        "install_dependencies": bool,
        "slack_webhook_url": str,
        "debug_time_enabled": bool,
        "debug_time": int,
        "email_on_error": bool,
        "email_on_success": bool,
        "queued_compute_wait_time": int,
        "idle_enabled": bool,
        "idle_time": int,
        "max_duration_workspaces_enabled": bool,
        "max_duration_workspaces": int,
        "max_duration_experiments_enabled": bool,
        "max_duration_experiments": int,
        "max_duration_endpoints_enabled": bool,
        "max_duration_endpoints": int,
        "max_duration_webapps_enabled": bool,
        "max_duration_webapps": int,
        "automatically_clear_cached_commits": int,
        "custom_pypi_enabled": bool,
        "custom_pypi_url": str,
        "default_compute_ids": list,
    }

    def __init__(self, organization):
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.ORGANIZATION_SETTINGS.format(scope["organization"])

        self._attributes = {}

    def save(self):
        """
        Save the local settings in the current organization
        @return: None
        """
        self.update(**self._attributes)

    def update(self, **kwargs):
        """
        Updates current organization's settings with the given params
        @param kwargs: any param out of the available attributes can be sent
        @return: None
        """

        def set_attributes_value(attributes):
            def set_value(key, value):
                if value is not None:
                    return value

                if OrganizationSettings.available_attributes[key] is str:
                    return DISABLED_PROPERTY_STRING

                if OrganizationSettings.available_attributes[key] is bool:
                    return bool(value)

                return DISABLED_PROPERTY_INT if OrganizationSettings.available_attributes[key] is int else value

            return {k: set_value(k, v) for k, v in attributes.items()}

        attributes_validator(
            available_attributes=OrganizationSettings.available_attributes,
            attributes=kwargs,
        )

        try:
            response = self._proxy.call_api(
                route=self._route,
                http_method=HTTP.PUT,
                payload=JAF.serialize(type="settings", attributes={**self._attributes, **set_attributes_value(kwargs)})
            )
        except CnvrgHttpError as e:
            self.reload()
            raise e

        self._attributes = response.attributes

    def values(self):
        hash_representation = {}
        for key in self.__class__.available_attributes.keys():
            hash_representation[key] = getattr(self, key)

        return hash_representation



import os
import random
import string
import sys

import pytest

# Add the parent folder to python path to run tests and coverage from shell
sys.path.append(os.path.join(os.getcwd(), os.pardir))


def random_string(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for _ in range(length))


@pytest.fixture(scope="function")
def random_name():
    return random_string


def get_domain():
    # Add env var CLUSTER_DOMAIN
    return os.environ.get("CLUSTER_DOMAIN") or "http://localhost:3000"


@pytest.fixture(scope="package")
def domain():
    return get_domain()


@pytest.fixture(scope="function", autouse=True)
def temp_file(tmpdir):
    file_path = "temp.txt"
    file = tmpdir.join(file_path)
    file.write("temp")
    return file



from cnvrgv2.config.error_messages import EMPTY_KAFKA_BROKERS_LIST, EMPTY_KAFKA_INPUT_TOPICS
from cnvrgv2.context import SCOPE
from cnvrgv2.config import routes
from cnvrgv2.errors import CnvrgArgumentsError

from cnvrgv2.modules.base.workflows_base import WorkflowsBase
from cnvrgv2.modules.workflows import Endpoint, EndpointKind, EndpointEnvSetup


class EndpointsClient(WorkflowsBase):
    def __init__(self, project):
        super().__init__(Endpoint, "Endpoint", project._context)

        scope = self._context.get_scope(SCOPE.PROJECT)
        self._route = routes.ENDPOINTS_BASE.format(scope["organization"], scope["project"])

    def create(
        self,
        title,
        file_name=None,
        function_name=None,
        kind=EndpointKind.WEB_SERVICE,
        env_setup=EndpointEnvSetup.PYTHON3,
        templates=None,
        kafka_brokers=None,
        kafka_input_topics=None,
        *args,
        **kwargs
    ):
        """
        Create a new endpoint

        @param title: Name of the endpoint
        @param templates: List of template names to be used
        @param kind: Integer representing the endpoints type. Use EndpointKind enum
        @param file_name: The file containing the endpoint's functions
        @param function_name: The name of the function the endpoint will route to
        @param env_setup: The interpreter to use. Use EndpointEnvSetup enum
        @param kafka_brokers: List of kafka brokers
        @param kafka_input_topics: List of topics to register as input
        @param args: optional arguments
        @param kwargs: Dictionary. Rest of optional attributes for creation
            image: Image object to create endpoint with
            queue: Name of the queue to run this job on
            kafka_output_topics: List of topics to register as output
            command_arguments: [Object] of key value arguments
            model_id: [String] name of the model
        TODO: Add a list of optional attributes
        @return: The newly created endpoint object
        """

        kwargs = {
            "kind": kind,
            "file_name": file_name,
            "function_name": function_name,
            "env_setup": env_setup,
            **kwargs
        }

        if kind == EndpointKind.STREAM:
            self._validate_stream_input(kafka_brokers, kafka_input_topics)
            kwargs["kafka_brokers"] = kafka_brokers
            kwargs["kafka_input_topics"] = kafka_input_topics

        return super().create(title, templates, *args, **kwargs)

    def _validate_stream_input(self, kafka_brokers, kafka_input_topics):
        error_dict = dict()
        if type(kafka_brokers) != list or len(kafka_brokers) < 1:
            error_dict[kafka_brokers] = EMPTY_KAFKA_BROKERS_LIST

        if type(kafka_input_topics) != list or len(kafka_input_topics) < 1:
            error_dict[kafka_input_topics] = EMPTY_KAFKA_INPUT_TOPICS

        if error_dict:
            raise CnvrgArgumentsError(error_dict)



from cnvrgv2.config.config import SSO_VERSION
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.modules.users.user import User
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgHttpError, CnvrgLoginError, CnvrgError
from cnvrgv2.config import routes, error_messages
from cnvrgv2.config.error_messages import FAULTY_VALUE
from cnvrgv2.utils.validators import validate_url, validate_email, validate_username


class UsersClient:
    def __init__(self, domain, token=None):
        if not validate_url(domain):
            raise CnvrgArgumentsError(error_messages.INVALID_URL)

        self._domain = domain
        self._token = token
        self._proxy = Proxy(domain=domain, token=token)
        self.sso_version = self._get_sso_version()
        self._proxy._sso_version = self.sso_version

    def _get_sso_version(self):
        """
        Find sso version based on version api response
        @return: sso version
        """
        response = self._proxy.call_api(
            route=routes.VERSION,
            http_method=HTTP.GET,
        )

        return response.attributes.get("sso_version", SSO_VERSION.CAPI)

    def login(self, user, password, token=None):
        """
        Authenticates the user with the given username/password
        @param user: The users Email
        @param password: the users password
        @raise CnvrgHttpError: If the user data is incorrect
        @return: token, first user organization (if exists) and boolean sso_enabled
        @param token: the users token
        """

        try:
            # If authentication fails, the proxy will throw unauthorized error
            response = self._proxy.call_api(
                route=routes.USER_LOGIN,
                http_method=HTTP.POST,
                payload={
                    "username": user,
                    "password": password,
                    "token": token
                },
            )

            token = response.meta.get("jwt")
            organization = response.meta.get("organization", None)
            sso_enabled = response.meta.get("sso_enabled", True)

        # update token and proxy for further usage
            self._token = token
            self._proxy = Proxy(domain=self._domain, token=self._token)

            return token, organization, sso_enabled
        except (CnvrgHttpError, CnvrgError):
            raise CnvrgLoginError(error_messages.INVALID_CREDENTIALS) from None

    def register(self, username, email, password):
        """
        Creates a new user using the provided argumnets
        @param username: The username
        @param email: The email
        @param password: The password
        @raise CnvrgHttpError: if the user already exists
        @return: True if the registration is successful
        """

        if not validate_email(email):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(email))

        if not validate_username(username):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(username))

        attributes = {
            "email": email,
            "username": username,
            "password": password
        }

        response = self._proxy.call_api(
            route=routes.USER_BASE,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="user", attributes=attributes)
        )

        return User(
            domain=self._domain,
            token=response.meta["jwt"],
            attributes=response.attributes
        )

    def me(self):
        """
        Retrieves current user information
        @raise CnvrgError: If the current context does not hold a user
        @return: Token, first user organization (if exists)
        """
        if self._token is None:
            raise CnvrgArgumentsError(error_messages.CONTEXT_CANT_SAVE)

        response = self._proxy.call_api(
            route=routes.USER_CURRENT,
            http_method=HTTP.GET
        )

        return User(domain=self._domain, token=self._token, attributes=response.attributes)



from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.resources.cluster import Cluster
from cnvrgv2.modules.resources.clusters_client import ClustersClient
from cnvrgv2.proxy import Proxy


class MockResponse:
    def __init__(self, status, json):
        self.status_code = status
        self._json = json

    def json(self):
        return self._json


class TestResource:
    def test_create_cluster_raise_onetype_error(self, mocker, empty_context, empty_cnvrg):

        empty_cnvrg._context = empty_context
        mocker.patch.object(Proxy, 'call_api', side_effect=MockResponse(status=200, json={}))
        client = ClustersClient(empty_cnvrg)
        try:
            client.create(build_yaml_path="fake1.yaml", kube_config_yaml_path="fake2.yaml")
        except CnvrgArgumentsError as err:
            assert "The given arguments: build_yaml_path and kube_config_yaml_path cannot be sent together" in str(err)

    def test_update_cluster_raise_NOFILE_error(self, mocker, empty_context, empty_cnvrg):

        empty_cnvrg._context = empty_context
        mocker.patch.object(Proxy, 'call_api', side_effect=MockResponse(status=200, json={}))
        cluster = Cluster(empty_cnvrg, slug="fake")
        try:
            cluster.update(kube_config_yaml_path="fake2.yaml")
        except CnvrgArgumentsError as err:
            assert "Bad arguments: File not found" in str(err)

    def test_delete_cluster_raise_slug_error(self, mocker, empty_context, empty_cnvrg):

        empty_cnvrg._context = empty_context
        client = ClustersClient(empty_cnvrg)
        try:
            client.delete(slug="")
        except CnvrgArgumentsError as err:
            assert "Bad arguments: Cannot get cluster with empty or a non string slug" in str(err)



import json
import os
import sys
import time
import requests
from datetime import datetime
from cnvrgv2.config import routes, error_messages
from cnvrgv2.config.error_messages import ARGUMENT_BAD_TYPE
from cnvrgv2.errors import CnvrgError, CnvrgArgumentsError
from cnvrgv2.modules.base.workflow_instance_base import WorkflowInstanceBase
from cnvrgv2.modules.images.image import Image
from cnvrgv2.modules.workflows.endpoint.endpoint_rule import EndpointRule
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.cron import Cron
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.log_utils import LOGS_TYPE_INFO, LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR, \
    LOGS_TYPE_WARNING
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.utils.time import get_relative_time


class EndpointKind:
    WEB_SERVICE = 0
    STREAM = 1
    BATCH = 2
    TGI = 7


class EndpointEnvSetup:
    PYTHON2 = "python_2"
    PYTHON3 = "python_3"
    PYSPARK = "pyspark"
    RENDPOINT = "r_endpoint"


class FeedbackLoopKind:
    IMMEDIATE = 0
    RECURRING = 1


class Endpoint(WorkflowInstanceBase):
    available_attributes = {
        "kind": str,
        "compute": str,
        "updated_at": str,
        "last_deployment": dict,
        "deployments": dict,
        "deployments_count": int,
        "templates": list,
        "endpoint_url": str,
        "current_deployment": dict,
        "compute_name": str,
        "image_name": str,
        "image_slug": str,
        "url": str,
        "api_key": str,
        "created_at": datetime,
        "max_replica": int,
        "min_replica": int,
        "export_data": bool,
        "conditions": dict,
        "full_href": str,
        **WorkflowInstanceBase.available_attributes
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.ENDPOINT, slug)

        self._scope = self._context.get_scope(SCOPE.ENDPOINT)

        self._proxy = Proxy(context=self._context)
        self._route = routes.ENDPOINT_BASE.format(
            self._scope["organization"],
            self._scope["project"],
            self._scope["endpoint"]
        )

        self._attributes = attributes or {}
        self._type = "Endpoint"
        self.slug = self._scope["endpoint"]

    def update_version(self, file_name=None, function_name=None, prep_file=None, prep_function=None, max_timeout=None,
                       **kwargs):
        """
        Update the endpoint version
        @param file_name: Name of new file
        @param function_name: Name of new function
        @param prep_file: Name of new preprocess file
        @param prep_function: Name of new preprocess function
        @param max_timeout: Time in minutes to wait until timeout
        @param kwargs:
        image: Image object to update endpoint with
        commit: Commit where files are.
        git_branch: Git branch where files are
        git_commit: Git commit where files are
        integer desired_percentage: Traffic ratio for canary rollout
        gunicorn_config: Array of key values in the following format: ["key=value", "key=value"]
        flask_config: Array of key values in the following format: ["key=value", "key=value"]
        kafka_brokers: List of kafka brokers
        kafka_input_topics: List of topics to register as input
        kafka_output_topics: List of topics to register as output
        input_file: Boolean. Does endpoint accepts file
        emails_to_notify: List. list of emails to send on success / error emails
        @return: updated endpoint
        """
        update_version_url = urljoin(self._route, routes.UPDATE_MODEL_VERSION)

        attributes = {
            "workflow_slug": self.slug,
            "file_name": file_name,
            "function_name": function_name,
            "prep_file": prep_file,
            "prep_function": prep_function,
            **kwargs
        }

        if isinstance(kwargs.get("image"), Image):
            attributes["image_slug"] = kwargs.get("image").slug

        start_time = datetime.now()

        res_attributes = self._proxy.call_api(
            route=update_version_url,
            http_method=HTTP.POST,
            payload=attributes
        ).attributes

        self.reload()
        while self.current_deployment["is_update"]:
            if max_timeout is not None:
                minutes_diff = (datetime.now() - start_time).total_seconds() / 60
                if minutes_diff > float(max_timeout):
                    raise CnvrgError(error_messages.ENDPOINT_UPDATE_TIMEOUT.format(self.slug))
            time.sleep(5)
            self.reload()

        self._attributes = {**self._attributes, **res_attributes}
        return self

    def rollback_version(self, version_slug=None):
        """
        Rollback an endpoint version
        @param version_slug: Version to rollback slug
        @return: None
        """
        rollback_version_url = urljoin(self._route, routes.ROLLBACK_MODEL_VERSION.format(version_slug))

        self._proxy.call_api(
            route=rollback_version_url,
            http_method=HTTP.POST,
        )

    def configure_feedback_loop(self, dataset_slug=None, scheduling_type=FeedbackLoopKind.IMMEDIATE, cron_string=None):
        """
        Update the endpoint version
        @param dataset_slug: Name of the dataset to export to
        @param Feedback loop kind: 0 - immediate export, 1 - recurring export
        @param cron_string: cron string if scheduling type is recurring
        @return: None
        """
        feedback_url = urljoin(self._route, routes.ENDPOINT_FEEDBACK_LOOP)
        if cron_string:
            Cron.validate_cron_syntax(cron_string)

        attributes = {
            "form_values": {
                "dataset": {
                    "slug": dataset_slug,
                },
                "sched_type": scheduling_type,
                "cron_string": cron_string
            }
        }

        self._proxy.call_api(
            route=feedback_url,
            http_method=HTTP.POST,
            payload=attributes
        )

        return

    def get_sample_code(self):
        """
        Get a sample code to query a running endpoint
        @return: Object with 2 properties: python & curl
        """

        route = urljoin(self._route, routes.ENDPOINT_SAMPLE_CODE)
        route = urljoin(self._route, "sample_code")
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.GET
        )

        return response.attributes

    def update_replicas(self, min_replica, max_replica):
        """
        Update endpoint number of replicas
        @param min_replica: minimum number of replicas
        @param max_replica: maximum number of replicas
        @return: None
        """
        self.reload()
        if self.status != 'ongoing':
            raise CnvrgError(error_messages.ENDPOINT_NOT_RUNNING.format(self.slug))

        route = urljoin(self._route, routes.ENDPOINT_REPLICAS)

        attributes = {
            "min_replica": min_replica,
            "max_replica": max_replica
        }

        self._proxy.call_api(
            route=route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def poll_charts(self):
        """
        Get endpoint charts and metrics
        @return: endpoint logs object with requests and latency properties
        """
        route = urljoin(self._route, routes.ENDPOINT_POLL_CHARTS)
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.GET
        )

        return response.attributes

    def stop_feedback_loop(self):
        """
        Stop existing endpoint feedback loop
        @return: None
        """
        route = urljoin(self._route, routes.ENDPOINT_FEEDBACK_LOOP)

        self._proxy.call_api(
            route=route,
            http_method=HTTP.DELETE
        )

    def batch_is_running(self):
        """
        Check whether batch endpoint is running or paused
        @return: True if running, false otherwise
        """
        route = urljoin(self._route, routes.ENDPOINT_STATUS)

        res = self._proxy.call_api(
            route=route,
            http_method=HTTP.GET
        )
        return res.meta['batch_is_running']

    def batch_scale_up(self):
        """
        Scale up batch endpoint
        @return: None
        """
        route = urljoin(self._route, routes.ENDPOINT_BATCH_SCALE_UP)

        self._proxy.call_api(
            route=route,
            http_method=HTTP.POST
        )

    def batch_scale_down(self):
        """
        Scale down batch endpoint
        @return: None
        """
        route = urljoin(self._route, routes.ENDPOINT_BATCH_SCALE_DOWN)

        self._proxy.call_api(
            route=route,
            http_method=HTTP.POST
        )

    def get_rules(self):
        """
        list endpoint rules
        @return: List of the endpoint rules (array of EndpointRule objects)
        """
        route = urljoin(self._route, routes.ENDPOINT_RULES)
        return api_list_generator(context=self._context, route=route, object=EndpointRule)

    def add_rule(
            self,
            title,
            metric,
            severity,
            operation,
            action,
            frequency=1,
            description=None,
            threshold=None,
            min_events=None
    ):
        """
        add a rule
        @param title: rule title
        @param description: rule description
        @param metric: The cnvrg SDK metric used for the trigger. (Only tags with numeric values are supported)
        @param threshold: The value you are comparing against.
        @param min_events: Minimum number of events before triggering action
        @param severity: An indication of the importance of this alert (Info, Warning or Critical).
        @param operation: The type of comparison used for comparing with your set value (greater than or less than).
        @param frequency: How often to run condition (in minutes)
        @param action: Action to occur when the alert is triggered (a json of: type, webhook slug, flow_slug and emails)
        @return: The new rule created
        """

        attributes = {
            "rule": {
                "title": title,
                "description": description,
                "metric": metric,
                "threshold": threshold,
                "min_events": min_events,
                "severity": severity.lower(),
                "operation": operation,
                "frequency": frequency,
                "action": {**action, "type": action.get("type", '').lower()}
            }
        }

        route = urljoin(self._route, routes.ENDPOINT_RULES)
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

        slug = response.attributes['slug']
        return EndpointRule(context=self._context, slug=slug, attributes=response.attributes)

    def update_rule(self, slug, is_active):
        """
        update rule state
        @param slug: the rule slug
        @param is_active: is the rule active or not
        """

        route = urljoin(self._route, routes.ENDPOINT_RULES + "/{0}".format(slug))
        rule = {
            "rule": {
                "is_active": is_active
            }
        }

        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=rule)
        )

        slug = response.attributes['slug']
        return EndpointRule(context=self._context, slug=slug, attributes=response.attributes)

    def delete_rule(self, slug):
        """
        delete a rule
        @param slug: rule slug
        """

        route = urljoin(self._route, routes.ENDPOINT_RULES + "/{0}".format(slug))
        self._proxy.call_api(
            route=route,
            http_method=HTTP.DELETE
        )

    def log_metric(self, name, y: float = None, x=None):
        params = {}
        try:
            params[name] = float(y)
            params["y"] = float(y)
        except Exception as e:
            print("ERROR {}".format(e), file=sys.stderr)
            return
        params["x"] = x
        params["cnvrg_metric_name"] = name
        log = {
            "job_id": self.slug,
            "job_type": self._type,
            "owner": self._context.get_scope(SCOPE.ENDPOINT)["organization"],
            "project": self._context.get_scope(SCOPE.ENDPOINT)["project"],
            "model": os.environ.get("CNVRG_MODEL_NAME", "unknown"),
            "event_time": time.time(),
            **params
        }
        print(json.dumps(log))
        sys.stdout.flush()

    def set_metadata(self, metadata=None):
        """
        Get a metadata of an endpoint
        @param metadata: Hash. metadata of the endpoint
        """
        route = urljoin(self._route, "metadata")
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.POST,
            payload={"metadata": metadata}
        )
        return response

    def get_metadata(self):
        """
        Get a metadata of an endpoint
        @return: Hash
        """
        route = urljoin(self._route, "metadata")
        response = self._proxy.call_api(
            route=route,
            http_method=HTTP.GET
        )

        return response.attributes.get("metadata")

    def predict(self, input_params):
        """
        Get a prediction from the endpoint's model
        @param input_params: the input parameters for the predict function
        @return: The prediction object
        """
        if not self.endpoint_url or not self.api_key:
            self.reload()
        headers = {
            'Cnvrg-Api-Key': self.api_key,
            'Content-Type': 'application/json',
        }
        data = json.dumps({"input_params": input_params})
        response = requests.post(self.endpoint_url, headers=headers,
                                 data=data)
        res_content = response.content.decode()

        print(res_content)
        return res_content

    def restart(self):
        """
        Override restart from workflows_base to remove functionality.
        restart() is only relevant for webapps
        """
        raise AttributeError("'Endpoint' object has no attribute 'restart'")

    def log(self, logs, log_level=LOGS_TYPE_OUTPUT):
        """
        Method to add logs to the endpoint log
        @param logs: an array of logs you want to send
        @param log_level: level of the logs, exists in log_utils
        """

        if log_level not in (LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR, LOGS_TYPE_INFO, LOGS_TYPE_WARNING):
            raise CnvrgArgumentsError({"level": ARGUMENT_BAD_TYPE.format("logs level enum", log_level)})

        if type(logs) is not list:
            if type(logs) is str:
                # give the user the option to only send one string
                logs = [logs]
            else:
                raise CnvrgArgumentsError({"level": ARGUMENT_BAD_TYPE.format("str", type(logs))})

        return super().write_logs(logs, log_level)

    def get_predictions(self, start_time=get_relative_time(months=6), end_time=get_relative_time(), offset=None,
                        model=None,
                        size=50):
        """
        Get last 50 predictions from the endpoint
        @param start_time: the start time of the predictions, default: 6 months ago
        @param end_time: the end time of the predictions, default: now
        @param size: the size of each batch of predictions
        @param offset: Offset pagination
        @param model: endpoint model number
        @return: The predictions JAF Attributes object
        """
        attributes = {
            "start_time": start_time,
            "end_time": end_time,
            "size": size,
        }
        if offset:
            attributes["offset"] = offset
        if model:
            attributes["model"] = model

        response = self._proxy.call_api(
            route=urljoin(self._route, routes.ENDPOINT_GET_PREDICTIONS),
            http_method=HTTP.GET,
            payload=attributes
        )

        return response.attributes




from cnvrgv2.config import routes
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import Proxy
from cnvrgv2.context import Context, SCOPE


class Volume(DynamicAttributes):

    available_attributes = {
        "slug": str,
        "title": str,
        "total_space": int,
        "used_space": int,
        "host_path": str,
        "host_ip": str,
        "read_only": bool,
        "hide": bool,
        "volume_type": str,
        "mount_path": str,
        "claim_name": str,
        "status": str,
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current volume
        if slug:
            self._context.set_scope(SCOPE.VOLUME, slug)

        scope = self._context.get_scope(SCOPE.VOLUME)

        self._proxy = Proxy(context=self._context)
        self._route = routes.VOLUME_BASE.format(scope["organization"], scope["project"], scope["volume"])
        self._attributes = attributes or {}
        self.slug = scope["volume"]



from tests.e2e.conftest import call_database


class DataOwnerLabels:
    @staticmethod
    def cleanup(context_prefix):
        delete_command = "DELETE FROM labels WHERE name LIKE'{}%'".format(context_prefix)
        call_database(delete_command)

    def data_owner_list_labels(self, e2e_data_owner, e2e_label):
        e2e_data_owner.labels.add(e2e_label)

        label = next(e2e_data_owner.labels.list())
        assert label.name == e2e_label.name
        assert label.color_name == e2e_label.color_name

    def data_owner_add_labels(self, e2e_data_owner, e2e_label):
        e2e_data_owner.labels.add(e2e_label)
        label = next(e2e_data_owner.labels.list())
        assert label.name == e2e_label.name
        assert label.color_name == e2e_label.color_name

    def data_owner_remove_labels(self, e2e_data_owner, e2e_label):
        e2e_data_owner.labels.add(e2e_label)
        label = next(e2e_data_owner.labels.list())
        assert label.name == e2e_label.name

        e2e_data_owner.labels.remove()

        labels = list(e2e_data_owner.labels.list())
        assert len(labels) == 0



from unittest.mock import MagicMock

import pytest
from cnvrgv2.modules.datasource import Datasource
from cnvrgv2.proxy import Proxy

bucket_files = [
    ['file1.txt', 'file2.txt'],
    ['file3.txt'],
    ['path2/'],
    ['path1/file1.txt', 'path1/file2.txt', 'path1/file3.txt']
]

bucket_files_without_path1 = [
    ['file1.txt', 'file2.txt'],
    ['file3.txt'],
    ['path2/'],
    ['file1.txt', 'file2.txt', 'file3.txt']
]
bucket_files_without_path2 = [
    ['file1.txt', 'file2.txt'],
    ['file3.txt'],
    [],
    ['path1/file1.txt', 'path1/file2.txt', 'path1/file3.txt']
]

file_path = 'my-file.txt'
normalized_file_name = 'normalized_path/' + file_path


# Pytest fixture to create the Datasource instance
@pytest.fixture(scope="function")
def datasource(empty_context, mocker):
    # Mock the API call globally for all tests
    mocker.patch.object(Proxy, 'call_api')
    ds = Datasource(context=empty_context, slug="test-slug")
    ds.name = 'test-name'
    ds.bucket_name = 'test-bucket-name'
    return ds


@pytest.fixture(scope="function")
def mock_datasource_client(mocker, datasource):
    """
    Fixture that patches the Datasource.get_s3_client method and returns the mock client.
    """
    client = MagicMock()
    mocker.patch.object(Datasource, 'get_s3_client', return_value=client)
    datasource._client = client
    return client


@pytest.fixture(scope="function")
def mock_datasource_files_downloader(mocker):
    mock_downloader = MagicMock()
    pages = [1, 2, 3]
    mock_downloader.page_iterator = pages
    return mocker.patch.object(Datasource, 'get_files_downloader', return_value=mock_downloader)


def s3_data():
    """
    Fixture that provides mock S3 data for testing.
    """
    data = []
    for files in bucket_files:
        obj = {'Contents': []}
        for file in files:
            obj['Contents'].append({'Key': file})
        data.append(obj)
    return data


@pytest.fixture(scope="function")
def mock_datasource_client_paginator(mocker, datasource, mock_datasource_client):
    mock_paginator = MagicMock()
    list_iter = s3_data()

    mock_paginator.paginate.return_value = iter(list_iter)
    mocker.patch.object(mock_datasource_client, 'get_paginator', return_value=mock_paginator)


@pytest.fixture(scope="function")
def mock_normalize_file_path(mocker):
    return mocker.patch.object(Datasource, '_normalize_file_path', return_value=normalized_file_name)



import pytest

from cnvrgv2.errors import CnvrgError


class TestProjectSettings:
    def test_get_settings_success(self, e2e_project):
        assert e2e_project.settings

    def test_save_settings_success(self, e2e_project):
        command = "git command"
        e2e_project.settings.command_to_execute = command
        e2e_project.settings.save()

        assert e2e_project.settings.command_to_execute == command

        e2e_project.settings.command_to_execute = "git command --other"
        e2e_project.settings.save()

        assert e2e_project.settings.command_to_execute != command

    def test_save_settings_default_image_success(self, e2e_project):
        default_image = "cnvrg_spark:v2.4.4-3.2"
        e2e_project.settings.default_image = default_image
        e2e_project.settings.save()
        assert e2e_project.settings.default_image == default_image

    def test_update_settings_success(self, e2e_project):
        default_computes = ["small", "large"]
        check_stuckiness = True
        stuck_time = 99

        e2e_project.settings.update(
            default_computes=default_computes,
            check_stuckiness=check_stuckiness,
            stuck_time=stuck_time
        )

        assert e2e_project.settings.default_computes == default_computes
        assert e2e_project.settings.check_stuckiness == check_stuckiness
        assert e2e_project.settings.stuck_time == stuck_time

    def test_update_non_existing_attribute_fail(self, e2e_project):
        with pytest.raises(CnvrgError):
            e2e_project.settings.update(fake="setting")



import os
import platform
import shutil
import time
from subprocess import PIPE, Popen
from urllib.parse import urlparse

import pytest
import requests
from requests.exceptions import ConnectionError

from cnvrgv2.cnvrg import Cnvrg
from cnvrgv2.config import CONFIG_FOLDER_NAME
from cnvrgv2.config import routes
from cnvrgv2.modules.organization.organizations_client import OrganizationsClient
from cnvrgv2.modules.users import ROLES
from cnvrgv2.modules.users.users_client import UsersClient
from cnvrgv2.modules.workflows import EndpointEnvSetup, EndpointKind, NotebookType, WebappType
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.modules.labels.utils import LabelColor, LabelKind
from tests.conftest import get_domain, random_string

datasource_test_bukcet_name = 'cnvrg-datasources-bucket'
datasource_test_region = 'us-east-2'


class TestContext(object):
    def __init__(self, prefix):
        self.prefix = prefix

    def generate_name(self, length):
        return self.prefix + random_string(length)


def call_database(command):
    username = os.environ.get("PSQL_USER")
    password = os.environ.get("PSQL_PASS")
    database_name = os.environ.get("PSQL_DATABASE", "cnvrg_development")  # Default name for localhost

    command = [get_psql_path(), "-h", get_postgres_hostname(), "-d", database_name, "-c", command]
    env_cp = os.environ.copy()

    if all([username, password]):
        command += ["-U", username]
        env_cp["PGPASSWORD"] = password

    pipe = Popen(command, stdout=PIPE, env=env_cp)
    result = pipe.communicate()
    return result[0].decode("utf-8")


def get_postgres_hostname():
    return urlparse(get_domain()).hostname.replace("app", "postgres")


def get_psql_path():
    if platform.system().lower() == "linux":  # This is when running automated tests on a linux machine
        psql_path = "/usr/bin/psql"
    else:  # And this is for running on developer's env
        psql_path = "/usr/local/bin/psql"  # This is intel chip
        if not os.path.isfile(psql_path):
            psql_path = "/opt/homebrew/bin/psql"  # And this is to support M1
    return psql_path


@pytest.fixture(autouse=True, scope="class")
def class_context(request):
    context_prefix = request.cls.__name__.lower() + random_string(3)

    yield TestContext(context_prefix)

    # Remove config file
    path_to_remove = CONFIG_FOLDER_NAME
    if os.path.exists(path_to_remove):
        shutil.rmtree(CONFIG_FOLDER_NAME)

    # Clean using class cleanup function
    cleanup = request.cls.__dict__.get("cleanup")
    if cleanup:
        cleanup_func = getattr(cleanup, '__func__')
        cleanup_func(context_prefix)


@pytest.fixture(scope="package")
def e2e_env(domain):
    # Wait till the server is up and running
    while True:
        try:
            resp = requests.get(urljoin(domain, "api", routes.VERSION))
            if resp.status_code == 200:
                break
            else:
                time.sleep(1)
        except ConnectionError:
            time.sleep(1)


@pytest.fixture(scope="class")
def e2e_user(e2e_env, domain, request):
    username = random_string(5)
    email = username + "@cnvrg.io"
    password = "qwe123"

    uc = UsersClient(domain=domain)
    uc.register(username=username, email=email, password=password)
    token, _, _ = uc.login(user=email, password=password)

    def cleanup():
        delete_user_command = "DELETE FROM users WHERE username = '{}'".format(username)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return {
        "token": token,
        "username": username,
        "email": email,
        "password": password,
    }


@pytest.fixture(scope="class")
def e2e_client(e2e_user, domain, request):
    org = random_string(5)

    oc = OrganizationsClient(domain=domain, token=e2e_user["token"])
    oc.create(name=org)

    cnvrg = Cnvrg(
        domain=domain,
        email=e2e_user["email"],
        password=e2e_user["password"],
        organization=org
    )

    # Update cluster's status to be online, so that templates could be used
    update_cluster_status = "UPDATE clusters SET status=2 WHERE id IN(SELECT max(id) FROM clusters)"
    call_database(update_cluster_status)

    # Update organization plan to premium to be able to add new clusters
    id_command = "SELECT id FROM organizations WHERE slug = '{}'".format(org)
    command_output = call_database(id_command)
    org_id = int(command_output.split('\n')[2].strip())

    # rails saves hashes as yaml
    metadata = "--- \n :unlimited: true \n :enterprise: true"
    update_plan = f"UPDATE plans SET metadata = '{metadata}' WHERE organization_id = {org_id}"
    call_database(update_plan)

    def cleanup():
        id_command = "SELECT id FROM organizations WHERE slug = '{}'".format(org)
        command_output = call_database(id_command)
        organization_id = int(command_output.split('\n')[2].strip())

        delete_clusters_command = "DELETE FROM clusters WHERE organization_id={}".format(organization_id)
        call_database(delete_clusters_command)

        delete_memebers_command = "DELETE FROM memberships WHERE organization_id={}".format(organization_id)
        call_database(delete_memebers_command)

        delete_images_command = "DELETE FROM images WHERE organization_id={}".format(organization_id)
        call_database(delete_images_command)

        delete_registries_command = "DELETE FROM registries WHERE organization_id={}".format(organization_id)
        call_database(delete_registries_command)

        delete_plans_command = "DELETE FROM plans WHERE organization_id={}".format(organization_id)
        call_database(delete_plans_command)

        delete_organization_command = "DELETE FROM organizations WHERE id={}".format(organization_id)
        call_database(delete_organization_command)

    request.addfinalizer(cleanup)

    return cnvrg


@pytest.fixture(scope="class")
def e2e_data_scientist_user(request, domain, e2e_client):
    username = random_string(5)
    email = username + "@cnvrg.io"
    password = "qwe123"

    uc = UsersClient(domain=domain)
    uc.register(username=username, email=email, password=password)
    e2e_client.members.add(email=email, role=ROLES.MEMBER)
    token, _, _ = uc.login(user=email, password=password)

    def cleanup():
        delete_user_command = "DELETE FROM users WHERE username = '{}'".format(username)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return {
        "token": token,
        "username": username,
        "email": email,
        "password": password,
    }


@pytest.fixture(scope="function")
def e2e_project(request, e2e_client, domain):
    project_name = random_string(5)
    project = e2e_client.projects.create(project_name)

    def cleanup():
        delete_project_command = "DELETE FROM projects WHERE title = '{}'".format(project_name)
        call_database(delete_project_command)

    request.addfinalizer(cleanup)

    return project


@pytest.fixture(scope="function")
def e2e_local_experiment(request, e2e_client, e2e_project, domain):
    experiment_name = random_string(5)
    experiment = e2e_project.experiments.init(experiment_name=experiment_name)
    slug = experiment.slug

    def cleanup():
        get_workflow_command = "SELECT id FROM workflows WHERE slug = '{}'".format(slug)
        command_output = call_database(get_workflow_command)
        organization_id = int(command_output.split('\n')[2].strip())
        delete_session_command = "DELETE FROM session_updates WHERE organization_id={}".format(organization_id)
        call_database(delete_session_command)
        delete_workflow_command = "DELETE FROM workflows WHERE slug = '{}'".format(slug)
        call_database(delete_workflow_command)

    request.addfinalizer(cleanup)

    return experiment


@pytest.fixture(scope="function")
def e2e_git_project(e2e_client, domain):
    project_name = random_string(5)
    project = e2e_client.projects.create(project_name)
    project.settings.update(**{
        "git_repo": "https://github.com/githubtraining/hellogitworld",
        "git_branch": "master",
        "is_git": True
    })
    return project


@pytest.fixture(scope="function", autouse=True)
def temp_file(tmpdir):
    file_path = "temp.txt"
    file = tmpdir.join(file_path)
    file.write("temp")
    return file


@pytest.fixture(scope="function", autouse=True)
def second_temp_file(tmpdir):
    file_path = "temp2.txt"
    file = tmpdir.join(file_path)
    file.write("temp 2")
    return file


@pytest.fixture(scope="function")
def e2e_flow(request, e2e_project):
    flow = e2e_project.flows.create()
    request.addfinalizer(lambda: flow.delete())
    return flow


@pytest.fixture(scope="class")
def e2e_flow_yaml():
    # TODO: Make it generate a yaml with a custom flow name or at least edit the file with a given flow name
    return {
        "title": "e2e Flow",
        "path": os.path.dirname(__file__) + "/assets/flow.yaml",
        "string": '''---
                    flow: e2e Flow
                    recurring:
                    next_run_utc:
                    tasks: []
                    relations: []'''
    }


@pytest.fixture(scope="function")
def e2e_flow_version(e2e_flow, domain):
    flow_version = next(e2e_flow.flow_versions.list())
    return flow_version


@pytest.fixture(scope="function")
def e2e_workspace(request, e2e_project, domain):
    title = random_string(5)
    templates = ["dev.small"]

    workspace = e2e_project.workspaces.create(
        title=title,
        templates=templates,
        notebook_type=NotebookType.JUPYTER_LAB
    )

    def cleanup():
        delete_user_command = "DELETE FROM workflows WHERE title = '{}'".format(title)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return workspace


@pytest.fixture(scope="function")
def e2e_webapp(request, e2e_project, domain):
    title = random_string(5)
    templates = ["dev.small"]
    non_existing_file = "filename.py"

    webapp = e2e_project.webapps.create(
        title=title,
        templates=templates,
        file_name=non_existing_file,
        webapp_type=WebappType.SHINY
    )

    def cleanup():
        delete_user_command = "DELETE FROM workflows WHERE title = '{}'".format(title)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return webapp


@pytest.fixture(scope="function")
def e2e_endpoint(request, e2e_project, domain):
    title = "ep-" + random_string(5)
    templates = ["dev.small"]
    non_existing_file = "filename.py"
    function_name = "predict"

    endpoint = e2e_project.endpoints.create(
        title=title,
        templates=templates,
        kind=EndpointKind.WEB_SERVICE,
        file_name=non_existing_file,
        function_name=function_name,
        env_setup=EndpointEnvSetup.PYTHON3
    )

    def cleanup():
        delete_user_command = "DELETE FROM workflows WHERE title = '{}'".format(title)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return endpoint


@pytest.fixture(scope="function")
def e2e_experiment(request, e2e_project):
    title = random_string(5)
    templates = ["dev.small"]
    command = "python3 -c 1+1"

    experiment = e2e_project.experiments.create(
        title=title,
        templates=templates,
        command=command,
        sync_before=False,
        sync_after=False,
    )

    def cleanup():
        delete_user_command = "DELETE FROM workflows WHERE title = '{}'".format(title)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return experiment


@pytest.fixture(scope="function")
def e2e_empty_experiment(request, e2e_project):
    title = random_string(5)

    def cleanup():
        delete_user_command = "DELETE FROM workflows WHERE title = '{}'".format(title)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return e2e_project.experiments.init(title)


@pytest.fixture(scope="function")
def e2e_dataset(request, e2e_client):
    dataset_name = random_string(5)
    dataset = e2e_client.datasets.create(dataset_name)

    request.addfinalizer(lambda: dataset.delete())

    return dataset


@pytest.fixture(scope="function")
def e2e_cluster(request, e2e_client):
    cluster_name = "cnvrg-test"
    attributes = {
        "domain": "test-domain.cicd.cnvrg.me",
        "scheduler": "cnvrg_scheduler",
        "namespace": "test-namespace"
    }
    cluster = e2e_client.clusters.create(
        kube_config_yaml_path="assets/kube_config.yaml",
        resource_name=cluster_name,
        domain=attributes["domain"],
        scheduler=attributes["scheduler"],
        namespace=attributes["namespace"],
        https_scheme=True,
        persistent_volumes=True,
        gaudi_enabled=True
    )

    def cleanup():
        delete_cluster_command = "DELETE FROM clusters WHERE title = '{}'".format(cluster_name)
        call_database(delete_cluster_command)

    request.addfinalizer(cleanup)

    return cluster


@pytest.fixture(scope="function")
def e2e_label_project(request, e2e_client):
    label_name = random_string(5)
    color_name = LabelColor.GREEN
    kind = LabelKind.PROJECT
    label = e2e_client.labels.create(name=label_name, kind=kind, color_name=color_name)

    def cleanup():
        delete_command = "DELETE FROM labels WHERE name = '{}'".format(label_name)
        call_database(delete_command)

    request.addfinalizer(cleanup)

    return label


@pytest.fixture(scope="function")
def e2e_label_dataset(request, e2e_client):
    label_name = random_string(5)
    color_name = LabelColor.RED
    kind = LabelKind.DATASET
    label = e2e_client.labels.create(name=label_name, kind=kind, color_name=color_name)

    def cleanup():
        delete_command = "DELETE FROM labels WHERE name = '{}'".format(label_name)
        call_database(delete_command)

    request.addfinalizer(cleanup)

    return label


@pytest.fixture(scope="function", autouse=True)
def temp_file_second(tmpdir):
    file_path = "temp2.txt"
    file = tmpdir.join(file_path)
    file.write("temp2")
    return file


@pytest.fixture(scope="function")
def e2e_temp_user(e2e_client, e2e_env, domain, request):
    username = random_string(5)
    email = username + "@cnvrg.io"
    password = "qwe123"

    uc = UsersClient(domain=domain)
    uc.register(username=username, email=email, password=password)
    e2e_client.members.add(email=email, role=ROLES.MEMBER)

    def cleanup():
        delete_user_command = "DELETE FROM users WHERE username = '{}'".format(username)
        call_database(delete_user_command)

    request.addfinalizer(cleanup)

    return {
        "username": username,
        "email": email,
        "password": password
    }



import pytest

from cnvrgv2.cnvrg import Cnvrg
from cnvrgv2.errors import CnvrgError, CnvrgHttpError


class TestOrganizationSettings:
    def test_get_settings_success(self, e2e_client):
        assert e2e_client.settings.default_computes
        assert e2e_client.settings.debug_time
        assert e2e_client.settings.max_duration_workspaces

    def test_save_settings_success(self, e2e_client):
        # Test with getattr (that fetches data from server) first
        assert e2e_client.settings.debug_time

        e2e_client.settings.debug_time = 50
        e2e_client.settings.save()

        e2e_client.settings._attributes = {}
        assert e2e_client.settings.debug_time == 50

        # Test without getattr (that fetches data from server) first
        e2e_client.settings._attributes = {}
        e2e_client.settings.debug_time = 60
        e2e_client.settings.save()

        e2e_client.settings._attributes = {}
        assert e2e_client.settings.debug_time == 60

    def test_update_settings_success(self, e2e_client):
        default_computes = ["small", "large"]
        idle_enabled = True
        idle_time = 99

        e2e_client.settings.update(
            default_computes=default_computes,
            idle_enabled=idle_enabled,
            idle_time=idle_time
        )

        assert e2e_client.settings.default_computes == default_computes
        assert e2e_client.settings.idle_enabled == idle_enabled
        assert e2e_client.settings.idle_time == idle_time

    def test_update_non_existing_attribute_fail(self, e2e_client):
        # Fetch original setting value for comparison
        original_setting = e2e_client.settings.email_on_error

        with pytest.raises(CnvrgError):
            e2e_client.settings.update(fake="setting", email_on_error=(not original_setting))

        assert e2e_client.settings.email_on_error == original_setting

    def test_update_unauthorized(self, domain, e2e_client, e2e_data_scientist_user):
        client_of_non_admin_user = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
            organization=e2e_client._context.organization
        )

        with pytest.raises(CnvrgHttpError) as exception_info:
            client_of_non_admin_user.settings.update(debug_time=666)

        assert 'unauthorized' in str(exception_info.value).lower()



import pytest

from cnvrgv2.errors import CnvrgError
from tests.e2e.conftest import call_database


class TestImages:
    @staticmethod
    def cleanup(context_prefix):
        delete_user_command = "DELETE FROM images WHERE name LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    def test_create_image(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        tag = "v1"
        registry = "docker-hub"

        image = e2e_client.images.create(name=name, tag=tag, registry=registry)
        assert image.slug
        assert image.tag == tag
        assert image.name == name
        assert image.dockerfile == ''
        assert image.registry_slug == registry

    def test_create_custom_image(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        tag = "v1"
        registry = "docker-hub"
        dockerfile = "docker"

        image = e2e_client.images.create(name=name, tag=tag, registry=registry, custom=True, dockerfile=dockerfile)
        assert image.slug
        assert image.tag == tag
        assert image.name == name
        assert image.dockerfile == dockerfile  # Tests dockerfile property as well!

    def test_get_image(self, class_context, e2e_client):
        name = "cnvrg"
        tag = "v5.0"

        # Get image by name and tag
        image = e2e_client.images.get(name="cnvrg", tag=tag)
        assert image.slug
        assert image.tag == tag
        assert image.name == name

        # Get image by slug
        image_by_slug = e2e_client.images.get(slug=image.slug)
        assert image_by_slug.tag == tag
        assert image_by_slug.name == name

    def test_update_image(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        tag = "v1"
        registry = "docker-hub"

        image = e2e_client.images.create(name=name, tag=tag, registry=registry)
        assert image.logo is None
        assert image.readme == ''

        logo = "cnvrg"
        readme = "readme"
        image.update(logo=logo, readme=readme)
        image.reload()
        assert image.logo == logo
        assert image.readme == readme

        image.update(logo=logo, readme=readme + "-new")
        image.reload()
        assert image.logo == logo
        assert image.readme == readme + "-new"

    def test_delete_image(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        tag = "v1"
        registry = "docker-hub"

        image = e2e_client.images.create(name=name, tag=tag, registry=registry)
        assert image.slug

        image.delete()
        with pytest.raises(CnvrgError):
            e2e_client.images.get(slug=image.slug).logo  # Need to call attribute to bypass lazy load

    def test_delete_image_from_client(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        tag = "v1"
        registry = "docker-hub"

        image = e2e_client.images.create(name=name, tag=tag, registry=registry)
        assert image.slug

        e2e_client.images.delete(slug=image.slug)
        with pytest.raises(CnvrgError):
            e2e_client.images.get(slug=image.slug).logo  # Need to call attribute to bypass lazy load
